#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import pytest

from PyInstaller import compat
from PyInstaller._shared_with_waf import _pyi_machine


def test_exec_command_subprocess_wrong_encoding_reports_nicely(capsys):
    # Ensure a nice error message is printed if decoding the output of the subprocess fails.
    # As `exec_python()` is used for running the program, we can use a small Python script.
    prog = """import sys; sys.stdout.buffer.write(b'dfadfadf\\xa0:::::')"""
    with pytest.raises(UnicodeDecodeError):
        compat.exec_python('-c', prog)
    out, err = capsys.readouterr()
    assert 'bytes around the offending' in err


# List every known platform.machine() or waf's ctx.env.DEST_CPU (in the bootloader/wscript file) output on Linux here.
@pytest.mark.parametrize(
    "input, output", [
        ("x86_64", "intel"),
        ("x64", "intel"),
        ("i686", "intel"),
        ("i386", "intel"),
        ("x86", "intel"),
        ("armv5", "arm"),
        ("armv7h", "arm"),
        ("armv7a", "arm"),
        ("arm", "arm"),
        ("aarch64", "arm"),
        ("ppc64le", "ppc"),
        ("ppc64", "ppc"),
        ("ppc32le", "ppc"),
        ("powerpc", "ppc"),
        ("s390x", "s390x"),
        ("mips", "mips"),
        ("mips64", "mips"),
        ("something-alien", "unknown"),
    ]
)
def test_linux_machine(input, output):
    assert _pyi_machine(input, "Linux") == output


def test_windows_machine():
    assert _pyi_machine("AMD64", "Windows") == "intel"
    assert _pyi_machine("ARM64", "Windows") == "arm"


def test_other_os_machine():
    assert _pyi_machine("foo", "Darwin") is None
    assert _pyi_machine("foo", "FreeBSD") is None



#-----------------------------------------------------------------------------
# Copyright (c) 2022-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
#
# Basic tests for macOS app bundle data relocation and code signing.

import os  # for `os.readlink`, because `pathlib.Path.readlink` is not available in python 3.8
import plistlib

import pytest

from PyInstaller.utils.tests import importorskip, onedir_only
from PyInstaller.building.osx import DOT_REPLACEMENT

# NOTE: the tests below explicitly enable the following environment variables:
#  PYINSTALLER_STRICT_BUNDLE_CODESIGN_ERROR: have codesign errors terminate the build instead of generating a warning.
#  PYINSTALLER_VERIFY_BUNDLE_SIGNATURE: perform strict verification of signature and signature-related data.
# We have these enabled globally on our CI (to cover all macOS .app bundle builds), but we explicitly apply them here
# for off-line test runs.


# Test that collected metadata is properly relocated to avoid codesign errors due to directory containing dots in name.
@pytest.mark.darwin
@importorskip('psutil')
@onedir_only
def test_macos_bundle_signing_metadata(pyi_builder, monkeypatch):
    monkeypatch.setenv("PYINSTALLER_STRICT_BUNDLE_CODESIGN_ERROR", "1")
    monkeypatch.setenv("PYINSTALLER_VERIFY_BUNDLE_SIGNATURE", "1")

    pyi_builder.test_source("""
        import psutil
        """, pyi_args=['--windowed', '--copy-metadata', 'psutil'])


# Test that the bundle signing works even if we collect a package as source .py files, which we do not relocate.
@pytest.mark.darwin
@importorskip('psutil')
@onedir_only
def test_macos_bundle_signing_py_files(pyi_builder, monkeypatch):
    monkeypatch.setenv("PYINSTALLER_STRICT_BUNDLE_CODESIGN_ERROR", "1")
    monkeypatch.setenv("PYINSTALLER_VERIFY_BUNDLE_SIGNATURE", "1")

    # Override Analysis so that we can set package collection mode without having to use .spec file.
    def AnalysisOverride(*args, **kwargs):
        kwargs['module_collection_mode'] = {'psutil': 'py'}
        return Analysis(*args, **kwargs)

    import PyInstaller.building.build_main
    Analysis = PyInstaller.building.build_main.Analysis
    monkeypatch.setattr('PyInstaller.building.build_main.Analysis', AnalysisOverride)

    pyi_builder.test_source("""
        import psutil
        """, pyi_args=['--windowed'])


# Test that the codesigning works even if we collect a package as .pyc files, which we do not relocate.
@pytest.mark.darwin
@importorskip('psutil')
@onedir_only
def test_macos_bundle_signing_pyc_files(pyi_builder, monkeypatch):
    monkeypatch.setenv("PYINSTALLER_STRICT_BUNDLE_CODESIGN_ERROR", "1")
    monkeypatch.setenv("PYINSTALLER_VERIFY_BUNDLE_SIGNATURE", "1")

    # Override Analysis so that we can set package collection mode without having to use .spec file.
    def AnalysisOverride(*args, **kwargs):
        kwargs['module_collection_mode'] = {'psutil': 'pyc'}
        return Analysis(*args, **kwargs)

    import PyInstaller.building.build_main
    Analysis = PyInstaller.building.build_main.Analysis
    monkeypatch.setattr('PyInstaller.building.build_main.Analysis', AnalysisOverride)

    pyi_builder.test_source("""
        import psutil
        """, pyi_args=['--windowed'])


# The following tests explicitly check the structure of generated macOS .app bundles w.r.t. binaries and data resources
# (re)location and cross-linking.


def _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=None, binaries=None):
    monkeypatch.setenv("PYINSTALLER_STRICT_BUNDLE_CODESIGN_ERROR", "1")
    monkeypatch.setenv("PYINSTALLER_VERIFY_BUNDLE_SIGNATURE", "1")

    extra_args = []
    for src_name, dest_name in datas or []:
        extra_args += ['--add-data', f"{src_name}:{dest_name}"]
    for src_name, dest_name in binaries or []:
        extra_args += ['--add-binary', f"{src_name}:{dest_name}"]

    pyi_builder.test_source("""
        print("Hello world!")
        """, pyi_args=['--windowed', *extra_args])

    # Return path to the generated .app bundle, so calling test can inspect it
    return tmp_path / 'dist' / 'test_source.app'


def _create_test_data_file(filename):
    filename.parent.mkdir(parents=True, exist_ok=True)
    # Create a text file
    filename.write_text("Test file", encoding='utf-8')


def _create_test_binary(filename):
    filename.parent.mkdir(parents=True, exist_ok=True)
    # Copy _struct extension
    import _struct
    import shutil
    shutil.copy2(_struct.__file__, filename)


def _create_test_framework(bundle_path):
    # Infer binary name from .framework bundle name.
    assert bundle_path.name.endswith('.framework')
    binary_name = bundle_path.name[:-10]

    binary_path = bundle_path / 'Versions' / 'A' / binary_name
    _create_test_binary(binary_path)

    # Create basic `Info.plist`.
    resources_path = bundle_path / 'Versions' / 'A' / 'Resources'
    resources_path.mkdir(parents=True, exist_ok=True)

    info_plist_file = resources_path / 'Info.plist'

    info_plist = {
        'CFBundleExecutable': binary_name,
        'CFBundleIdentifier': f'org.pyinstaller.{binary_name}',
        'CFBundlePackageType': 'FMWK',
        'CFBundleShortVersionString': '1.0',
        'CFBundleSignature': '????',
        'CFBundleVersion': '1.0.0',
    }

    with open(info_plist_file, "wb") as fp:
        plistlib.dump(info_plist, fp)


# Test that top-level data file is relocated into `Contents/Resources` and symlinked back into `Contents/Frameworks`.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_data_file(pyi_builder, monkeypatch, tmp_path):
    datas = []

    src_path = tmp_path / 'data_file.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, '.'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas)

    # The data file is placed into `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into `Contents/MacOS`.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'data_file.txt'
    assert check_path.is_symlink()
    assert check_path.is_file()  # Is link valid?
    assert os.readlink(check_path) == '../Resources/data_file.txt'


# Test that top-level binary is kept in `Contents/Frameworks` and symlinked into `Contents/Resources`.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_binary(pyi_builder, monkeypatch, tmp_path):
    binaries = []

    src_path = tmp_path / 'binary.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, '.'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, binaries=binaries)

    # The binary is placed into `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into `Contents/Resources`.
    check_path = bundle_path / 'Contents' / 'Resources' / 'binary.dylib'
    assert check_path.is_symlink()
    assert check_path.is_file()  # Is link valid?
    assert os.readlink(check_path) == '../Frameworks/binary.dylib'


# Test that data-only directory is relocated into `Contents/Resources` and symlinked back into `Contents/Frameworks`.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_data_only_dir(pyi_builder, monkeypatch, tmp_path):
    datas = []

    # + data_dir: (1)
    #    - data_file1.txt (2)
    #    - data_file2.txt (3)
    src_path = tmp_path / 'data_dir' / 'data_file1.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'data_dir'))

    src_path = tmp_path / 'data_dir' / 'data_file2.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'data_dir'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas)

    # (1) The whole data directory is placed into `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'data_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... and symlinked (at directory level) into `Contents/Frameworks`.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'data_dir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../Resources/data_dir'

    # (2) The data file is placed into directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'data_dir' / 'data_file1.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... but it is also reachable from `Contents/Frameworks`. The linking is done at the parent directory level, so
    # the file itself is NOT seen as a symlink.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'data_dir' / 'data_file1.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (3) Same goes for the second data file.
    check_path = bundle_path / 'Contents' / 'Resources' / 'data_dir' / 'data_file2.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Frameworks' / 'data_dir' / 'data_file2.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()


# Test that binary-only directory is kept in `Contents/Frameworks` and symlinked into `Contents/Resources`.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_binary_only_dir(pyi_builder, monkeypatch, tmp_path):
    binaries = []

    # + binary_dir: (1)
    #    - binary1.dylib (2)
    #    - binary2.dylib (3)
    src_path = tmp_path / 'binary_dir' / 'binary1.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'binary_dir'))

    src_path = tmp_path / 'binary_dir' / 'binary2.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'binary_dir'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, binaries=binaries)

    # (1) The whole binary directory is placed into `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... and symlinked (at directory level) into Contents/Resources.
    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../Frameworks/binary_dir'

    # (2) The binary file is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'binary1.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... but it is also reachable from `Resources/Frameworks`. The linking is done at the parent directory level,
    # so the file itself is NOT seen as a symlink.
    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'binary1.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (3) Same goes for the second binary file.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'binary2.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'binary2.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()


# Test that mxied-content directory is created in both `Contents/Frameworks` and `Contents/Resources`, and that files
# are put into the proper directory and cross-linked into the other directory.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_mixed_dir(pyi_builder, monkeypatch, tmp_path):
    datas = []
    binaries = []

    # + mixed_dir: (1)
    #    - data_file.txt (2)
    #    - binary.dylib (3)
    src_path = tmp_path / 'mixed_dir' / 'data_file.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'mixed_dir'))

    src_path = tmp_path / 'mixed_dir' / 'binary.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'mixed_dir'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas, binaries=binaries)

    # (1) The mixed-content directory is created in both `Contents/Frameworks` and `Contents/Resources` (i.e., no
    # linking at the directory level).
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (2) The data file is placed into directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into directory in `Contents/Frameworks`.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'data_file.txt'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Resources/mixed_dir/data_file.txt'

    # (3) The binary is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into directory in `Contents/Resources`.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'binary.dylib'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Frameworks/mixed_dir/binary.dylib'


# Repeat the test with mixed-content directory, except that it now contains three sub-directories: a data-only one,
# a binary-only one, and mixed-content one.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_mixed_dir_with_subdirs(pyi_builder, monkeypatch, tmp_path):
    datas = []
    binaries = []

    # + mixed_dir: (1)
    #    + data_subdir: (2)
    #       - data_file.txt (3)
    #    + binary_subdir: (4)
    #       - binary.dylib (5)
    #    + mixed_subdir: (6)
    #       - data_file.txt (7)
    #       - binary.dylib (8)
    src_path = tmp_path / 'mixed_dir' / 'data_subdir' / 'data_file.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'mixed_dir/data_subdir'))

    src_path = tmp_path / 'mixed_dir' / 'binary_subdir' / 'binary.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'mixed_dir/binary_subdir'))

    src_path = tmp_path / 'mixed_dir' / 'mixed_subdir' / 'data_file.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'mixed_dir/mixed_subdir'))

    src_path = tmp_path / 'mixed_dir' / 'mixed_subdir' / 'binary.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'mixed_dir/mixed_subdir'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas, binaries=binaries)

    # (1) The mixed-content directory is created in both `Contents/Frameworkds` and `Contents/Resources` (i.e., no
    # linking at the directory level).
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (2) The whole data sub-directory is placed into directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'data_subdir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... and symlinked (at directory level) into directory in `Contents/Frameworks`.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'data_subdir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Resources/mixed_dir/data_subdir'

    # (3) The data file is placed into data sub-directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'data_subdir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... but it is also reachable from `Contents/Frameworks`. The linking is done at the parent directory level, so
    # the file itself is NOT seen as a symlink.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'data_subdir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (4) The whole binary sub-directory is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'binary_subdir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... and symlinked (at directory level) into directory in `Contents/Resources`.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'binary_subdir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Frameworks/mixed_dir/binary_subdir'

    # (5) The binary file is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'binary_subdir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... but it is also reachable from `Resources/Frameworks`. The linking is done at the parent directory level,
    # so the file itself is NOT seen as a symlink.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'binary_subdir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (6) The mixed-content sub-directory is created in both `Contents/Frameworks` and `Contents/Resources` (i.e., no
    # linking at the directory level).
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'mixed_subdir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'mixed_subdir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (7) The data file is placed into directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'mixed_subdir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into directory in `Contents/Frameworks`.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'mixed_subdir' / 'data_file.txt'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../../Resources/mixed_dir/mixed_subdir/data_file.txt'

    # (3) The binary is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'mixed_subdir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into directory in `Contents/Resources`.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'mixed_subdir' / 'binary.dylib'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../../Frameworks/mixed_dir/mixed_subdir/binary.dylib'


# Repeat the test with mixed-content directory and sub-directories, except that all directories now contain a dot in
# their names.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_mixed_dir_with_subdirs_and_dots(pyi_builder, monkeypatch, tmp_path):
    datas = []
    binaries = []

    # + mixed.dir: (1)
    #    + .data_subdir: (2)
    #       - data_file.txt (3)
    #    + .binary_subdir: (4)
    #       - binary.dylib (5)
    #    + mixed_subdir.: (6)
    #       - data_file.txt (7)
    #       - binary.dylib (8)
    src_path = tmp_path / 'mixed.dir' / '.data_subdir' / 'data_file.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'mixed.dir/.data_subdir'))

    src_path = tmp_path / 'mixed.dir' / '.binary_subdir' / 'binary.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'mixed.dir/.binary_subdir'))

    src_path = tmp_path / 'mixed.dir' / 'mixed_subdir.' / 'data_file.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'mixed.dir/mixed_subdir.'))

    src_path = tmp_path / 'mixed.dir' / 'mixed_subdir.' / 'binary.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'mixed.dir/mixed_subdir.'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas, binaries=binaries)

    # (1) The mixed-content directory is created in both `Contents/Frameworks` and `Contents/Resources` (i.e., no
    # linking at the directory level). For directory in `Contents/MacOS`, the `.` in the name is replaced with
    # `DOT_REPLACEMENT`, and a symlink is created from original name to the modified one. For directory in
    # `Contents/Resources`, this is not necessary.
    check_path = bundle_path / 'Contents' / 'Frameworks' / f'mixed{DOT_REPLACEMENT}dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == f'mixed{DOT_REPLACEMENT}dir'

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / f'mixed{DOT_REPLACEMENT}dir'
    assert not check_path.exists()

    # (2) The whole data sub-directory is placed into directory in `Contents/Resources`. There is no need for name
    # moficiation here.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / '.data_subdir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / f'{DOT_REPLACEMENT}data_subdir'
    assert not check_path.exists()

    # The data sub-directory is symlinked (at directory level) into directory in `Contents/Frameworks`. Due to this
    # being a symlink, we do not need name modification here, either.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / '.data_subdir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Resources/mixed.dir/.data_subdir'

    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / f'{DOT_REPLACEMENT}data_subdir'
    assert not check_path.exists()

    # (3) The data file is placed into data sub-directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / '.data_subdir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... but it is also reachable from `Contents/Frameworks`. The linking is done at the parent directory level,
    # so the file itself is NOT seen as a symlink.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / '.data_subdir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (4) The whole binary sub-directory is placed into directory in `Contents/Frameworks`. Due to the dot in the name,
    # the directory is created with modified name (`.` replaced by `DOT_REPLACEMENT`). A symlink is created from
    # original name to the modified one.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / f'{DOT_REPLACEMENT}binary_subdir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / '.binary_subdir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == f'{DOT_REPLACEMENT}binary_subdir'

    # A symlink (at directory level) is also created into directory in `Contents/Resources`.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / '.binary_subdir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Frameworks/mixed.dir/.binary_subdir'

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / f'{DOT_REPLACEMENT}binary_subdir'
    assert not check_path.exists()

    # (5) The binary file is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / '.binary_subdir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... but it is also reachable from `Resources/Frameworks`. The linking is done at the parent directory level,
    # so the file itself is NOT seen as a symlink.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / '.binary_subdir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (6) The mixed-content sub-directory is created in both `Contents/Frameworks` and `Contents/Resources` (i.e., no
    # linking at the directory level). The directory in `Contents/MacOS` requires special handling due to dot in the
    # name (replacement of `.` with `DOT_REPLACEMENT`, and symlink from original to modified name), while the directory
    # in `Contents/Resources` does not.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / f'mixed_subdir{DOT_REPLACEMENT}'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / 'mixed_subdir.'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == f'mixed_subdir{DOT_REPLACEMENT}'

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / 'mixed_subdir.'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / f'mixed_subdir{DOT_REPLACEMENT}'
    assert not check_path.exists()

    # (7) The data file is placed into directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / 'mixed_subdir.' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into directory in `Contents/Frameworks`.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / 'mixed_subdir.' / 'data_file.txt'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../../Resources/mixed.dir/mixed_subdir./data_file.txt'

    # (3) The binary is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed.dir' / 'mixed_subdir.' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into directory in `Contents/Resources`.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed.dir' / 'mixed_subdir.' / 'binary.dylib'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../../Frameworks/mixed.dir/mixed_subdir./binary.dylib'


# Test with symlink in top-level directory pointing to a data file in data-only directory.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_symlink_into_data_dir(pyi_builder, monkeypatch, tmp_path):
    datas = []

    # + data_dir: (1)
    #    - data_file.txt (2)
    # - link_to_data.txt -> data_dir/data_file.txt (3)
    src_path = tmp_path / 'data_dir' / 'data_file.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'data_dir'))

    src_path = tmp_path / 'link_to_data_file.txt'
    src_path.symlink_to('data_dir/data_file.txt')
    datas.append((src_path, '.'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas)

    # (1) The whole data directory is placed into `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'data_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... and symlinked (at directory level) into `Contents/Frameworks`.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'data_dir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../Resources/data_dir'

    # (2) The data file is placed into directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'data_dir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... but it is also reachable from `Contents/Frameworks`. The linking is done at the parent directory level,
    # so the file itself is NOT seen as a symlink.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'data_dir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (3) The symlink is replicated in both `Contents/Frameworks` and `Contents/Resources`, and points to the resource
    # (file or symlink) in the same directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'link_to_data_file.txt'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'data_dir/data_file.txt'

    check_path = bundle_path / 'Contents' / 'Resources' / 'link_to_data_file.txt'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'data_dir/data_file.txt'


# Test with symlink in top-level directory pointing to a binary in binary-only directory.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_symlink_into_binary_dir(pyi_builder, monkeypatch, tmp_path):
    binaries = []

    # + binary_dir: (1)
    #    - binary.dylib (2)
    # - link_to_binary.dylib -> binary_dir/binary.dylib (3)
    src_path = tmp_path / 'binary_dir' / 'binary.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'binary_dir'))

    src_path = tmp_path / 'link_to_binary.dylib'
    src_path.symlink_to('binary_dir/binary.dylib')
    binaries.append((src_path, '.'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, binaries=binaries)

    # (1) The whole binary directory is placed into `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... and symlinked (at directory level) into Contents/Resources.
    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../Frameworks/binary_dir'

    # (2) The binary file is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... but it is also reachable from `Resources/Frameworks`. The linking is done at the parent directory level,
    # so the file itself is NOT seen as a symlink.
    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (3) The symlink is replicated in both `Contents/Frameworks` and `Contents/Resources`, and points to the resource
    # (file or symlink) in the same directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'link_to_binary.dylib'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'binary_dir/binary.dylib'

    check_path = bundle_path / 'Contents' / 'Resources' / 'link_to_binary.dylib'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'binary_dir/binary.dylib'


# Test with symlinks in top-level directory pointing to files in mixed-content directory.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_symlink_into_mixed_dir(pyi_builder, monkeypatch, tmp_path):
    datas = []
    binaries = []

    # + mixed_dir: (1)
    #    - data_file.txt (2)
    #    - binary.dylib (3)
    # - link_to_data_file.txt -> mixed_dir/data_file.txt (4)
    # - link_to_binary.dylib -> mixed_dir/link_to_binary.dylib (5)
    src_path = tmp_path / 'mixed_dir' / 'data_file.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'mixed_dir'))

    src_path = tmp_path / 'mixed_dir' / 'binary.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'mixed_dir'))

    src_path = tmp_path / 'link_to_data_file.txt'
    src_path.symlink_to('mixed_dir/data_file.txt')
    datas.append((src_path, '.'))

    src_path = tmp_path / 'link_to_binary.dylib'
    src_path.symlink_to('mixed_dir/binary.dylib')
    binaries.append((src_path, '.'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas, binaries=binaries)

    # (1) The mixed-content directory is created in both `Contents/Frameworks` and `Contents/Resources` (i.e., no
    # linking at the directory level).
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (2) The data file is placed into directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into directory in `Contents/Frameworks`.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'data_file.txt'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Resources/mixed_dir/data_file.txt'

    # (3) The binary is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into directory in `Contents/Resources`.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'binary.dylib'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Frameworks/mixed_dir/binary.dylib'

    # (4) The symlink is replicated in both `Contents/Frameworks` and `Contents/Resources`, and points to the resource
    # (file or symlink) in the same directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'link_to_data_file.txt'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'mixed_dir/data_file.txt'

    check_path = bundle_path / 'Contents' / 'Resources' / 'link_to_data_file.txt'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'mixed_dir/data_file.txt'

    # (5) The symlink is replicated in both `Contents/Frameworks` and `Contents/Resources`, and points to the resource
    # (file or symlink) in the same directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'link_to_binary.dylib'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'mixed_dir/binary.dylib'

    check_path = bundle_path / 'Contents' / 'Resources' / 'link_to_binary.dylib'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'mixed_dir/binary.dylib'


# Test with .framework bundle in top-level directory and framework's binary symlinked to top-level directory.
# This implicitly also tests that we do not replace the dot in the .framework bundle's directory name (the .framework
# bundle directories are the only directories in `Contents/Frameworks` that are allowed to have a dot in name).
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_framework_in_top_level(pyi_builder, monkeypatch, tmp_path):
    datas = []
    binaries = []

    # + Dummy.framework: (1)
    #    + Versions: (2)
    #       + A: (3)
    #          - Dummy (4)
    #          + Resources: (5)
    #             - Info.plist (6)
    # - Dummy -> Dummy.framework/Versions/A/Dummy (7)

    src_path = tmp_path / 'Dummy.framework'
    _create_test_framework(src_path)
    binaries.append((src_path / 'Versions' / 'A' / 'Dummy', 'Dummy.framework/Versions/A'))
    # NOTE1: Info.plist should be collected automatically based on the collected framework binary.
    # NOTE2: symlink `Current` -> `A` should be automatically generated in `Dummy.framework/Versions`.

    src_path = tmp_path / 'Dummy'
    src_path.symlink_to('Dummy.framework/Versions/A/Dummy')
    binaries.append((src_path, '.'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas, binaries=binaries)

    # (1) The .framework directory is placed into `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy.framework'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... and symlinked (at directory level) into Contents/Resources.
    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy.framework'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../Frameworks/Dummy.framework'

    # (2) The content of .framework are placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy.framework' / 'Versions'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... but are also reachable from `Resources/Frameworks`. The linking is done at the parent directory level, so the
    # files/directories themselves are NOT seen as symlinks.
    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy.framework' / 'Versions'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (3) Same for `Versions/<version>` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy.framework' / 'Versions' / 'A'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy.framework' / 'Versions' / 'A'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (4) Same for binary within the `<version>` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy.framework' / 'Versions' / 'A' / 'Dummy'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy.framework' / 'Versions' / 'A' / 'Dummy'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (5) Same for `Resources` directory within the `<version>` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy.framework' / 'Versions' / 'A' / 'Resources'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy.framework' / 'Versions' / 'A' / 'Resources'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (6) Same for `Info.plist` in `Resources` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy.framework' / 'Versions' / 'A' / 'Resources' / \
        'Info.plist'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy.framework' / 'Versions' / 'A' / 'Resources' / \
        'Info.plist'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (X) A symlink `Current` pointing to `<version>` should be automatically created inside `Versions` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy.framework' / 'Versions' / 'Current'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'A'

    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy.framework' / 'Versions' / 'Current'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'A'

    # (7) The symlink is replicated in both `Contents/Frameworks` and `Contents/Resources`, and points to the resource
    # (file or symlink) in the same directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'Dummy.framework/Versions/A/Dummy'

    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'Dummy.framework/Versions/A/Dummy'


# Test with .framework bundle in binary-only directory and framework's binary symlinked to top-level directory.
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_framework_in_binary_dir(pyi_builder, monkeypatch, tmp_path):
    datas = []
    binaries = []

    # + binary_dir: (1)
    #    - binary.dylib (2)
    #    + Dummy.framework: (3)
    #       + Versions: (4)
    #          + A: (5)
    #             - Dummy (6)
    #             + Resources: (7)
    #                - Info.plist (8)
    # - Dummy -> binary_dir/Dummy.framework/Versions/A/Dummy (9)

    src_path = tmp_path / 'binary_dir' / 'binary.dylib'
    _create_test_binary(src_path)
    binaries.append((src_path, 'binary_dir'))

    src_path = tmp_path / 'binary_dir' / 'Dummy.framework'
    _create_test_framework(src_path)
    binaries.append((src_path / 'Versions' / 'A' / 'Dummy', 'binary_dir/Dummy.framework/Versions/A'))
    # NOTE1: Info.plist should be collected automatically based on the collected framework binary.
    # NOTE2: symlink `Current` -> `A` should be automatically generated in `Dummy.framework/Versions`.

    src_path = tmp_path / 'Dummy'
    src_path.symlink_to('binary_dir/Dummy.framework/Versions/A/Dummy')
    binaries.append((src_path, '.'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas, binaries=binaries)

    # (1) The whole binary directory is placed into `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... and symlinked (at directory level) into Contents/Resources.
    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../Frameworks/binary_dir'

    # (2) The binary file is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... but it is also reachable from `Resources/Frameworks`. The linking is done at the parent directory level,
    # so the file itself is NOT seen as a symlink.
    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'binary.dylib'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (3) The same applies to the .framework directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'Dummy.framework'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'Dummy.framework'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (4) Same for `Versions` directory inside .framework directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'Dummy.framework' / 'Versions'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'Dummy.framework' / 'Versions'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (5) Same for `Versions/<version>` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'A'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'A'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (6) Same for binary within the `<version>` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'A' / 'Dummy'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'A' / 'Dummy'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (7) Same for `Resources` directory within the `<version>` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'A' / \
        'Resources'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'A' / \
        'Resources'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (8) Same for `Info.plist` in `Resources` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'A' / \
        'Resources' / 'Info.plist'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'A' / \
        'Resources' / 'Info.plist'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (X) A symlink `Current` pointing to `<version>` should be automatically created inside `Versions` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'Current'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'A'

    check_path = bundle_path / 'Contents' / 'Resources' / 'binary_dir' / 'Dummy.framework' / 'Versions' / 'Current'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'A'

    # (9) The symlink is replicated in both `Contents/Frameworks` and `Contents/Resources`, and points to the resource
    # (file or symlink) in the same directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'binary_dir/Dummy.framework/Versions/A/Dummy'

    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'binary_dir/Dummy.framework/Versions/A/Dummy'


# Test with .framework bundle in mixed-content directory and framework's binary symlinked to top-level directory.
# This also tests that a directory is recognized as a mixed-content one if it contains a data file and a .framework
# bundle (i.e., no other binary files).
@pytest.mark.darwin
@onedir_only
def test_macos_bundle_layout_framework_in_mixed_dir(pyi_builder, monkeypatch, tmp_path):
    datas = []
    binaries = []

    # + mixed_dir: (1)
    #    - data_file.txt (2)
    #    + Dummy.framework: (3)
    #       + Versions: (4)
    #          + A: (5)
    #             - Dummy (6)
    #             + Resources: (7)
    #                - Info.plist (8)
    # - Dummy -> mixed_dir/Dummy.framework/Versions/A/Dummy (9)

    src_path = tmp_path / 'mixed_dir' / 'data_file.txt'
    _create_test_data_file(src_path)
    datas.append((src_path, 'mixed_dir'))

    src_path = tmp_path / 'mixed_dir' / 'Dummy.framework'
    _create_test_framework(src_path)
    binaries.append((src_path / 'Versions' / 'A' / 'Dummy', 'mixed_dir/Dummy.framework/Versions/A'))
    # NOTE1: Info.plist should be collected automatically based on the collected framework binary.
    # NOTE2: symlink `Current` -> `A` should be automatically generated in `Dummy.framework/Versions`.

    src_path = tmp_path / 'Dummy'
    src_path.symlink_to('mixed_dir/Dummy.framework/Versions/A/Dummy')
    binaries.append((src_path, '.'))

    bundle_path = _create_app_bundle(pyi_builder, monkeypatch, tmp_path, datas=datas, binaries=binaries)

    # (1) The mixed-content directory is created in both `Contents/Frameworks` and `Contents/Resources` (i.e., no
    # linking at the directory level).
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (2) The data file is placed into directory in `Contents/Resources`...
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'data_file.txt'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # ... and symlinked into directory in `Contents/Frameworks`.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'data_file.txt'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Resources/mixed_dir/data_file.txt'

    # (3) The .framework bundle directory is placed into directory in `Contents/Frameworks`...
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'Dummy.framework'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # ... and symlinked into `Resources/Frameworks`.
    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'Dummy.framework'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == '../../Frameworks/mixed_dir/Dummy.framework'

    # (4) The contents inside the .framework directory is reachable from both directories, and due to symlinking at
    # parent (= .framework) directory, they are not visible as symlinks themselves. In this case, this applies to the
    # `Versions` directory inside .framework directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'Dummy.framework' / 'Versions'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'Dummy.framework' / 'Versions'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (5) Same for `Versions/<version>` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'A'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'A'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (6) Same for binary within the `<version>` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'A' / 'Dummy'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'A' / 'Dummy'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (7) Same for `Resources` directory within the `<version>` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'A' / \
        'Resources'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'A' / \
        'Resources'
    assert check_path.is_dir()
    assert not check_path.is_symlink()

    # (8) Same for `Info.plist` in `Resources` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'A' / \
        'Resources' / 'Info.plist'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'A' / \
        'Resources' / 'Info.plist'
    assert check_path.is_file()
    assert not check_path.is_symlink()

    # (X) A symlink `Current` pointing to `<version>` should be automatically created inside `Versions` directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'Current'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'A'

    check_path = bundle_path / 'Contents' / 'Resources' / 'mixed_dir' / 'Dummy.framework' / 'Versions' / 'Current'
    assert check_path.is_dir()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'A'

    # (9) The symlink is replicated in both `Contents/Frameworks` and `Contents/Resources`, and points to the resource
    # (file or symlink) in the same directory.
    check_path = bundle_path / 'Contents' / 'Frameworks' / 'Dummy'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'mixed_dir/Dummy.framework/Versions/A/Dummy'

    check_path = bundle_path / 'Contents' / 'Resources' / 'Dummy'
    assert check_path.is_file()
    assert check_path.is_symlink()
    assert os.readlink(check_path) == 'mixed_dir/Dummy.framework/Versions/A/Dummy'



#-----------------------------------------------------------------------------
# Copyright (c) 2015-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# __import__ below works
from pyi_testmod_metapath1.extern import aaa  # noqa: F401
# __import__ below works
#import pyi_testmod_metapath1.extern
# __import__ below fails!
#import pyi_testmod_metapath1.extern.aaa
__import__('pyi_testmod_metapath1.extern.bbb')

try:
    bbb = pyi_testmod_metapath1.extern.bbb
except NameError:
    pass
else:
    print("name:", bbb.__name__)
    print("__file__:", bbb.__file__ or repr(bbb.__file__))
    assert bbb.__name__.endswith('._vendor.bbb')

# Second test: import sub-sub-package
# Mimic this lines from pkg_resources.__init__.py as of setuptools 28.6.1:
#   from pkg_resources.extern import packaging
#   __import__('pkg_resources.extern.packaging.version')
from pyi_testmod_metapath1.extern import ccc  # noqa: F401, E402

__import__('pyi_testmod_metapath1.extern.ccc.ddd')

try:
    ddd = pyi_testmod_metapath1.extern.ccc.ddd
except NameError:
    pass
else:
    print("name:", ddd.__name__)
    print("__file__:", ddd.__file__ or repr(ddd.__file__))
    assert ddd.__name__.endswith('._vendor.ccc.ddd')

# Third test: import sub-sub-sub-package
__import__('pyi_testmod_metapath1.extern.ccc.eee.fff')



from . import submodule_feature3

# Try importing two other custom top-level modules, and provide boolean flags indicating their availability.
# Feature #1
feature1_available = False
try:
    import mymodule_feature1  # noqa: F401
    feature1_available = True
except ImportError:
    pass

# Feature #2
feature2_available = False
try:
    import mymodule_feature2  # noqa: F401
    feature2_available = True
except ImportError:
    pass

# Feature #3: imported via sub-module from this pacakge, which in turn tries to import submodule from mymodule_feature3.
feature3_available = submodule_feature3.feature3_available



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
In `#4592 <https://github.com/pyinstaller/pyinstaller/pull/4592>`_, the bootloader was modified to allow the execution
traceback to be shown in Windowed mode. Unfortunately, this modification caused the recurrence of a bug raised in #1869.
"""

import pytest


@pytest.mark.parametrize(
    'src,retcode',
    [
        # Code to run, retcode
        ('raise SystemExit', 0),
        ('import sys; sys.exit()', 0),
        ('raise SystemExit(1)', 1),
        ('import sys; sys.exit(2)', 2),
        ('raise SystemExit("Message to get printed to the console.")', 1),
        ('raise Exception("Unhandled exception.")', 1)  # See issue #5480
    ]
)
def test_systemexit_is_handled_correctly(src, retcode, pyi_builder):
    pyi_builder.test_source(src, retcode=retcode)



import sys
import unittest
import textwrap
import subprocess
import os
from PyInstaller.lib.modulegraph import modulegraph

class TestNativeImport (unittest.TestCase):
    # The tests check that Python's import statement
    # works as these tests expect.

    def importModule(self, name):
        if '.' in name:
            script = textwrap.dedent("""\
                try:
                    import %s
                except ImportError:
                    import %s
                print (%s.__name__)
            """) %(name, name.rsplit('.', 1)[0], name)
        else:
            script = textwrap.dedent("""\
                import %s
                print (%s.__name__)
            """) %(name, name)

        p = subprocess.Popen([sys.executable, '-c', script],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-import-from-init'),
            encoding='utf8',
        )
        data = p.communicate()[0]
        data = data.strip()

        if data.endswith(' refs]'):
            # with --with-pydebug builds
            data = data.rsplit('\n', 1)[0].strip()

        sts = p.wait()

        if sts != 0:
            print (data)
        self.assertEqual(sts, 0)
        return data

    def testRootPkgRelImport(self):
        m = self.importModule('pkg2')
        self.assertEqual(m, 'pkg2')

    def testSubPackageRelImport(self):
        m = self.importModule('pkg2.subpkg')
        self.assertEqual(m, 'pkg2.subpkg')


class TestModuleGraphImport (unittest.TestCase):
    if not hasattr(unittest.TestCase, 'assertIsInstance'):
        def assertIsInstance(self, value, types):
            if not isinstance(value, types):
                self.fail("%r is not an instance of %r"%(value, types))

    def setUp(self):
        root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-import-from-init')
        self.mf = modulegraph.ModuleGraph(path=[ root ] + sys.path)
        #self.mf.debug = 999
        self.mf.add_script(os.path.join(root, 'script.py'))

    def testRootPkgRelImport(self):
        node = self.mf.find_node('pkg2')
        self.assertIsInstance(node, modulegraph.Package)
        self.assertEqual(node.identifier, 'pkg2')

    def testSubPackageRelImport(self):
        node = self.mf.find_node('pkg2.subpkg')
        self.assertIsInstance(node, modulegraph.Package)
        self.assertEqual(node.identifier, 'pkg2.subpkg')

        node = self.mf.find_node('pkg2.subpkg.compat')
        self.assertIsInstance(node, modulegraph.SourceModule)
        self.assertEqual(node.identifier, 'pkg2.subpkg.compat')

        node = self.mf.find_node('pkg2.subpkg._collections')
        self.assertIsInstance(node, modulegraph.SourceModule)
        self.assertEqual(node.identifier, 'pkg2.subpkg._collections')


if __name__ == "__main__":
    unittest.main()



# -----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# -----------------------------------------------------------------------------

# An importlib.util.LazyLoader test, based on example from
# https://docs.python.org/3/library/importlib.html#implementing-lazy-imports

import sys
import importlib.util


def lazy_import(name):
    spec = importlib.util.find_spec(name)
    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    return module


# Lazy-load the module...
lazy_module = lazy_import(sys.argv[1])
# ... and then trigger load by listing its contents
print(dir(lazy_module))



import unittest

import os, shutil

from PyInstaller.lib.modulegraph import modulegraph

class DummyModule(object):
    packagepath = None
    def __init__(self, ppath):
        self.packagepath = ppath

class FindAllSubmodulesTestCase(unittest.TestCase):
    def testNone(self):
        mg = modulegraph.ModuleGraph()
        # empty packagepath
        m = DummyModule(None)
        sub_ms = []
        for sm in mg._find_all_submodules(m):
            sub_ms.append(sm)
        self.assertEqual(sub_ms, [])

    def testSimple(self):
        mg = modulegraph.ModuleGraph()
        # a string does not break anything although it is split into its characters
        # BUG: "/hi/there" will read "/"
        m = DummyModule("xyz")
        sub_ms = []
        for sm in mg._find_all_submodules(m):
            sub_ms.append(sm)
        self.assertEqual(sub_ms, [])

    def testSlashes(self):
        # a string does not break anything although it is split into its characters
        # BUG: "/xyz" will read "/" so this one already triggers missing itertools
        mg = modulegraph.ModuleGraph()
        m = DummyModule("/xyz")
        sub_ms = []
        for sm in mg._find_all_submodules(m):
            sub_ms.append(sm)
        self.assertEqual(sub_ms, [])

if __name__ == '__main__':
    unittest.main()



"""
Test importability of SWIG-generated C extensions.
"""
import os
import sys
import unittest

from PyInstaller.lib.modulegraph import modulegraph


class TestSWIGImportability(unittest.TestCase):
    if not hasattr(unittest.TestCase, 'assertIsInstance'):
        def assertIsInstance(self, value, types):
            if not isinstance(value, types):
                self.fail("%r is not an instance of %r"%(value, types))

    def test_swig_importability(self):
        # Absolute path of the top-level data directory for this unit test.
        test_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), 'testpkg-swig')

        # Mock module graph relative to this directory.
        module_graph = modulegraph.ModuleGraph(path=[ test_dir ] + sys.path)

        # Graph node corresponding to a mock SWIG module.
        swig_module = module_graph._safe_import_hook(
            'pkg.sample', source_module=None, target_attr_names=())[0]
        self.assertIsInstance(swig_module, modulegraph.SourceModule)

        # Graph node corresponding to a mock SWIG C extension imported by the
        # prior module. While this should technically be a C extension rather
        # than a module, reliably testing the latter in a cross-platform manner
        # is both non-trivial and gains us relatively little over this approach.
        swig_c_extension = module_graph.find_node('pkg._sample')
        self.assertIsInstance(swig_c_extension, modulegraph.SourceModule)

if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from __future__ import with_statement
import os, shlex, shutil, traceback, errno, sys, stat
from waflib import Utils, Configure, Logs, Options, ConfigSet, Context, Errors, Build, Node

build_dir_override = None
no_climb_commands = ['configure']
default_cmd = "build"


def waf_entry_point(current_directory, version, wafdir):
    Logs.init_log()
    if Context.WAFVERSION != version:
        Logs.error('Waf script %r and library %r do not match (directory %r)', version, Context.WAFVERSION, wafdir)
        sys.exit(1)
    Context.waf_dir = wafdir
    Context.run_dir = Context.launch_dir = current_directory
    start_dir = current_directory
    no_climb = os.environ.get('NOCLIMB')
    if len(sys.argv) > 1:
        potential_wscript = os.path.join(current_directory, sys.argv[1])
        if os.path.basename(potential_wscript) == Context.WSCRIPT_FILE and os.path.isfile(potential_wscript):
            path = os.path.normpath(os.path.dirname(potential_wscript))
            start_dir = os.path.abspath(path)
            no_climb = True
            sys.argv.pop(1)
    ctx = Context.create_context('options')
    (options, commands, env) = ctx.parse_cmd_args(allow_unknown=True)
    if options.top:
        start_dir = Context.run_dir = Context.top_dir = options.top
        no_climb = True
    if options.out:
        Context.out_dir = options.out
    if not no_climb:
        for k in no_climb_commands:
            for y in commands:
                if y.startswith(k):
                    no_climb = True
                    break
    cur = start_dir
    while cur:
        try:
            lst = os.listdir(cur)
        except OSError:
            lst = []
            Logs.error('Directory %r is unreadable!', cur)
        if Options.lockfile in lst:
            env = ConfigSet.ConfigSet()
            try:
                env.load(os.path.join(cur, Options.lockfile))
                ino = os.stat(cur)[stat.ST_INO]
            except EnvironmentError:
                pass
            else:
                for x in (env.run_dir, env.top_dir, env.out_dir):
                    if not x:
                        continue
                    if Utils.is_win32:
                        if cur == x:
                            load = True
                            break
                    else:
                        try:
                            ino2 = os.stat(x)[stat.ST_INO]
                        except OSError:
                            pass
                        else:
                            if ino == ino2:
                                load = True
                                break
                else:
                    Logs.warn('invalid lock file in %s', cur)
                    load = False
                if load:
                    Context.run_dir = env.run_dir
                    Context.top_dir = env.top_dir
                    Context.out_dir = env.out_dir
                    break
        if not Context.run_dir:
            if Context.WSCRIPT_FILE in lst:
                Context.run_dir = cur
        next = os.path.dirname(cur)
        if next == cur:
            break
        cur = next
        if no_climb:
            break
    wscript = os.path.normpath(os.path.join(Context.run_dir, Context.WSCRIPT_FILE))
    if not os.path.exists(wscript):
        if options.whelp:
            Logs.warn('These are the generic options (no wscript/project found)')
            ctx.parser.print_help()
            sys.exit(0)
        Logs.error(
            'Waf: Run from a folder containing a %r file (or try -h for the generic options)', Context.WSCRIPT_FILE
        )
        sys.exit(1)
    try:
        os.chdir(Context.run_dir)
    except OSError:
        Logs.error('Waf: The folder %r is unreadable', Context.run_dir)
        sys.exit(1)
    try:
        set_main_module(wscript)
    except Errors.WafError as e:
        Logs.pprint('RED', e.verbose_msg)
        Logs.error(str(e))
        sys.exit(1)
    except Exception as e:
        Logs.error('Waf: The wscript in %r is unreadable', Context.run_dir)
        traceback.print_exc(file=sys.stdout)
        sys.exit(2)
    if options.profile:
        import cProfile, pstats
        cProfile.runctx('from waflib import Scripting; Scripting.run_commands()', {}, {}, 'profi.txt')
        p = pstats.Stats('profi.txt')
        p.sort_stats('time').print_stats(75)
    else:
        try:
            try:
                run_commands()
            except:
                if options.pdb:
                    import pdb
                    type, value, tb = sys.exc_info()
                    traceback.print_exc()
                    pdb.post_mortem(tb)
                else:
                    raise
        except Errors.WafError as e:
            if Logs.verbose > 1:
                Logs.pprint('RED', e.verbose_msg)
            Logs.error(e.msg)
            sys.exit(1)
        except SystemExit:
            raise
        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            sys.exit(2)
        except KeyboardInterrupt:
            Logs.pprint('RED', 'Interrupted')
            sys.exit(68)


def set_main_module(file_path):
    Context.g_module = Context.load_module(file_path)
    Context.g_module.root_path = file_path

    def set_def(obj):
        name = obj.__name__
        if not name in Context.g_module.__dict__:
            setattr(Context.g_module, name, obj)

    for k in (dist, distclean, distcheck):
        set_def(k)
    if not 'init' in Context.g_module.__dict__:
        Context.g_module.init = Utils.nada
    if not 'shutdown' in Context.g_module.__dict__:
        Context.g_module.shutdown = Utils.nada
    if not 'options' in Context.g_module.__dict__:
        Context.g_module.options = Utils.nada


def parse_options():
    ctx = Context.create_context('options')
    ctx.execute()
    if not Options.commands:
        if isinstance(default_cmd, list):
            Options.commands.extend(default_cmd)
        else:
            Options.commands.append(default_cmd)
    if Options.options.whelp:
        ctx.parser.print_help()
        sys.exit(0)


def run_command(cmd_name):
    ctx = Context.create_context(cmd_name)
    ctx.log_timer = Utils.Timer()
    ctx.options = Options.options
    ctx.cmd = cmd_name
    try:
        ctx.execute()
    finally:
        ctx.finalize()
    return ctx


def run_commands():
    parse_options()
    run_command('init')
    while Options.commands:
        cmd_name = Options.commands.pop(0)
        ctx = run_command(cmd_name)
        Logs.info('%r finished successfully (%s)', cmd_name, ctx.log_timer)
    run_command('shutdown')


def distclean_dir(dirname):
    for (root, dirs, files) in os.walk(dirname):
        for f in files:
            if f.endswith(('.o', '.moc', '.exe')):
                fname = os.path.join(root, f)
                try:
                    os.remove(fname)
                except OSError:
                    Logs.warn('Could not remove %r', fname)
    for x in (Context.DBFILE, 'config.log'):
        try:
            os.remove(x)
        except OSError:
            pass
    try:
        shutil.rmtree(Build.CACHE_DIR)
    except OSError:
        pass


def distclean(ctx):
    '''removes build folders and data'''
    def remove_and_log(k, fun):
        try:
            fun(k)
        except EnvironmentError as e:
            if e.errno != errno.ENOENT:
                Logs.warn('Could not remove %r', k)

    if not Options.commands:
        for k in os.listdir('.'):
            for x in '.waf-2 waf-2 .waf3-2 waf3-2'.split():
                if k.startswith(x):
                    remove_and_log(k, shutil.rmtree)
    cur = '.'
    if os.environ.get('NO_LOCK_IN_TOP') or ctx.options.no_lock_in_top:
        cur = ctx.options.out
    try:
        lst = os.listdir(cur)
    except OSError:
        Logs.warn('Could not read %r', cur)
        return
    if Options.lockfile in lst:
        f = os.path.join(cur, Options.lockfile)
        try:
            env = ConfigSet.ConfigSet(f)
        except EnvironmentError:
            Logs.warn('Could not read %r', f)
            return
        if not env.out_dir or not env.top_dir:
            Logs.warn('Invalid lock file %r', f)
            return
        if env.out_dir == env.top_dir:
            distclean_dir(env.out_dir)
        else:
            remove_and_log(env.out_dir, shutil.rmtree)
        env_dirs = [env.out_dir]
        if not (os.environ.get('NO_LOCK_IN_TOP') or ctx.options.no_lock_in_top):
            env_dirs.append(env.top_dir)
        if not (os.environ.get('NO_LOCK_IN_RUN') or ctx.options.no_lock_in_run):
            env_dirs.append(env.run_dir)
        for k in env_dirs:
            p = os.path.join(k, Options.lockfile)
            remove_and_log(p, os.remove)


class Dist(Context.Context):
    '''creates an archive containing the project source code'''
    cmd = 'dist'
    fun = 'dist'
    algo = 'tar.bz2'
    ext_algo = {}

    def execute(self):
        self.recurse([os.path.dirname(Context.g_module.root_path)])
        self.archive()

    def archive(self):
        import tarfile
        arch_name = self.get_arch_name()
        try:
            self.base_path
        except AttributeError:
            self.base_path = self.path
        node = self.base_path.make_node(arch_name)
        try:
            node.delete()
        except OSError:
            pass
        files = self.get_files()
        if self.algo.startswith('tar.'):
            tar = tarfile.open(node.abspath(), 'w:' + self.algo.replace('tar.', ''))
            for x in files:
                self.add_tar_file(x, tar)
            tar.close()
        elif self.algo == 'zip':
            import zipfile
            zip = zipfile.ZipFile(node.abspath(), 'w', compression=zipfile.ZIP_DEFLATED)
            for x in files:
                archive_name = self.get_base_name() + '/' + x.path_from(self.base_path)
                zip.write(x.abspath(), archive_name, zipfile.ZIP_DEFLATED)
            zip.close()
        else:
            self.fatal('Valid algo types are tar.bz2, tar.gz, tar.xz or zip')
        try:
            from hashlib import sha256
        except ImportError:
            digest = ''
        else:
            digest = ' (sha256=%r)' % sha256(node.read(flags='rb')).hexdigest()
        Logs.info('New archive created: %s%s', self.arch_name, digest)

    def get_tar_path(self, node):
        return node.abspath()

    def add_tar_file(self, x, tar):
        p = self.get_tar_path(x)
        tinfo = tar.gettarinfo(name=p, arcname=self.get_tar_prefix() + '/' + x.path_from(self.base_path))
        tinfo.uid = 0
        tinfo.gid = 0
        tinfo.uname = 'root'
        tinfo.gname = 'root'
        if os.path.isfile(p):
            with open(p, 'rb') as f:
                tar.addfile(tinfo, fileobj=f)
        else:
            tar.addfile(tinfo)

    def get_tar_prefix(self):
        try:
            return self.tar_prefix
        except AttributeError:
            return self.get_base_name()

    def get_arch_name(self):
        try:
            self.arch_name
        except AttributeError:
            self.arch_name = self.get_base_name() + '.' + self.ext_algo.get(self.algo, self.algo)
        return self.arch_name

    def get_base_name(self):
        try:
            self.base_name
        except AttributeError:
            appname = getattr(Context.g_module, Context.APPNAME, 'noname')
            version = getattr(Context.g_module, Context.VERSION, '1.0')
            self.base_name = appname + '-' + version
        return self.base_name

    def get_excl(self):
        try:
            return self.excl
        except AttributeError:
            self.excl = Node.exclude_regs + ' **/waf-2.* **/.waf-2.* **/waf3-2.* **/.waf3-2.* **/*~ **/*.rej **/*.orig **/*.pyc **/*.pyo **/*.bak **/*.swp **/.lock-w*'
            if Context.out_dir:
                nd = self.root.find_node(Context.out_dir)
                if nd:
                    self.excl += ' ' + nd.path_from(self.base_path)
            return self.excl

    def get_files(self):
        try:
            files = self.files
        except AttributeError:
            files = self.base_path.ant_glob('**/*', excl=self.get_excl())
        return files


def dist(ctx):
    '''makes a tarball for redistributing the sources'''
    pass


class DistCheck(Dist):
    fun = 'distcheck'
    cmd = 'distcheck'

    def execute(self):
        self.recurse([os.path.dirname(Context.g_module.root_path)])
        self.archive()
        self.check()

    def make_distcheck_cmd(self, tmpdir):
        cfg = []
        if Options.options.distcheck_args:
            cfg = shlex.split(Options.options.distcheck_args)
        else:
            cfg = [x for x in sys.argv if x.startswith('-')]
        cmd = [sys.executable, sys.argv[0], 'configure', 'build', 'install', 'uninstall', '--destdir=' + tmpdir] + cfg
        return cmd

    def check(self):
        import tempfile, tarfile
        with tarfile.open(self.get_arch_name()) as t:
            for x in t:
                t.extract(x)
        instdir = tempfile.mkdtemp('.inst', self.get_base_name())
        cmd = self.make_distcheck_cmd(instdir)
        ret = Utils.subprocess.Popen(cmd, cwd=self.get_base_name()).wait()
        if ret:
            raise Errors.WafError('distcheck failed with code %r' % ret)
        if os.path.exists(instdir):
            raise Errors.WafError('distcheck succeeded, but files were left in %s' % instdir)
        shutil.rmtree(self.get_base_name())


def distcheck(ctx):
    '''checks if the project compiles (tarball from 'dist')'''
    pass


def autoconfigure(execute_method):
    def execute(self):
        if not Configure.autoconfig:
            return execute_method(self)
        env = ConfigSet.ConfigSet()
        do_config = False
        try:
            env.load(os.path.join(Context.top_dir, Options.lockfile))
        except EnvironmentError:
            Logs.warn('Configuring the project')
            do_config = True
        else:
            if env.run_dir != Context.run_dir:
                do_config = True
            else:
                h = 0
                for f in env.files:
                    try:
                        h = Utils.h_list((h, Utils.readf(f, 'rb')))
                    except EnvironmentError:
                        do_config = True
                        break
                else:
                    do_config = h != env.hash
        if do_config:
            cmd = env.config_cmd or 'configure'
            if Configure.autoconfig == 'clobber':
                tmp = Options.options.__dict__
                launch_dir_tmp = Context.launch_dir
                if env.options:
                    Options.options.__dict__ = env.options
                Context.launch_dir = env.launch_dir
                try:
                    run_command(cmd)
                finally:
                    Options.options.__dict__ = tmp
                    Context.launch_dir = launch_dir_tmp
            else:
                run_command(cmd)
            run_command(self.cmd)
        else:
            return execute_method(self)

    return execute


Build.BuildContext.execute = autoconfigure(Build.BuildContext.execute)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import traceback, sys


class WafError(Exception):
    def __init__(self, msg='', ex=None):
        Exception.__init__(self)
        self.msg = msg
        assert not isinstance(msg, Exception)
        self.stack = []
        if ex:
            if not msg:
                self.msg = str(ex)
            if isinstance(ex, WafError):
                self.stack = ex.stack
            else:
                self.stack = traceback.extract_tb(sys.exc_info()[2])
        self.stack += traceback.extract_stack()[:-1]
        self.verbose_msg = ''.join(traceback.format_list(self.stack))

    def __str__(self):
        return str(self.msg)


class BuildError(WafError):
    def __init__(self, error_tasks=[]):
        self.tasks = error_tasks
        WafError.__init__(self, self.format_error())

    def format_error(self):
        lst = ['Build failed']
        for tsk in self.tasks:
            txt = tsk.format_error()
            if txt:
                lst.append(txt)
        return '\n'.join(lst)


class ConfigurationError(WafError):
    pass


class TaskRescan(WafError):
    pass


class TaskNotReady(WafError):
    pass



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller import compat
from PyInstaller.utils.hooks.setuptools import setuptools_info

datas = []

hiddenimports = [
    # Test case import/test_zipimport2 fails during importing pkg_resources or setuptools when module not present.
    'distutils.command.build_ext',
    'setuptools.msvc',
]

# Necessary for setuptools on Mac/Unix
if compat.is_unix or compat.is_darwin:
    hiddenimports.append('syslog')

# Prevent the following modules from being collected solely due to reference from anywhere within setuptools (or
# its vendored dependencies).
excludedimports = [
    'pytest',
    'numpy',  # originally from hook-setuptools.msvc
    'docutils',  # originally from hool-setuptools._distutils.command.check
]

# setuptools >= 39.0.0 is "vendoring" its own direct dependencies from "_vendor" to "extern". This also requires
# 'pre_safe_import_module/hook-setuptools.extern.six.moves.py' to make the moves defined in 'setuptools._vendor.six'
# importable under 'setuptools.extern.six'.
#
# With setuptools 71.0.0, the vendored packages are exposed to the outside world by `setuptools._vendor` location being
# appended to `sys.path`, and the `VendorImporter` is gone (i.e., no more mapping to `setuptools.extern`). Since the
# vendored dependencies are now exposed as top-level modules (provided upstream versions are not available, as they
# would take precedence due to `sys.path` ordering), we need pre-safe-import-module hooks that detect when only vendored
# version is available, and add aliases to prevent duplicated collection. For list of vendored packages for which we
# need such pre-safe-import-module hooks, see the code in `PyInstaller.utils.hooks.setuptools`.
#
# The list of submodules from `setuptools._vendor` is now available in `setuptools_info.vendored_modules` (and covers
# all setuptools versions).
#
# NOTE: with setuptools >= 71.0, we do not need to add modules from `setuptools._vendored` to hidden imports anymore,
# because the aliases we set up should ensure that the necessary parts get collected. We still need them for earlier
# versions of setuptools, though.
if setuptools_info.version < (71, 0):
    hiddenimports += setuptools_info.vendored_modules

# The situation with vendored distutils (from `setuptools._distutils`) is a bit more complicated; python >= 3.12 does
# not provide stdlib version of `distutils` anymore, so our corresponding pre-safe-import-module hook sets up aliases.
# In earlier python versions, stdlib version is available as well, and at run-time, we might need both versions present,
# so that whichever is applicable can be used. Therefore, for python < 3.12, we need to add the vendored distuils
# modules to hidden imports.
if setuptools_info.distutils_vendored and not compat.is_py312:
    hiddenimports += setuptools_info.distutils_modules

# With setuptools >= 71.0.0, the vendored packages also have metadata, and might also contain data files that need to
# be collected. The list of corresponding data files is kept cached in `setuptools_info.vendored_data` (to minimize the
# number of times we need to call collect_data_files()).
#
# While it might be tempting to simply collect all data files and be done with it, we actually need to match the
# collection behavior for the stand-alone versions of these packages; i.e., we should collect metadata (and/or data
# files) for the vendored package only if the same data is also collected for stand-alone version. Otherwise, we risk
# inconsistent behavior and potential mismatches; for example, if we collected metadata for vendored package A here,
# but end up collecting stand-alone A, for which we normally do not collect the metadata, then at run-time, we will end
# up with stand-alone copy of A and vendored copy of its metadata being discoverable.
#
# Therefore, if metadata and/or metadata needs to be collected, do it in corresponding sub-package hook (for an example,
# see `hook-setuptools._vendor.jaraco.text.py`).



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Find external dependencies of binary libraries.
"""

import ctypes.util
import functools
import os
import pathlib
import re
import sys
import sysconfig
import subprocess

from PyInstaller import compat
from PyInstaller import log as logging
from PyInstaller.depend import dylib, utils
from PyInstaller.utils.win32 import winutils
from PyInstaller.exceptions import PythonLibraryNotFoundError

if compat.is_darwin:
    import PyInstaller.utils.osx as osxutils

logger = logging.getLogger(__name__)

_exe_machine_type = None
if compat.is_win:
    _exe_machine_type = winutils.get_pe_file_machine_type(compat.python_executable)

#- High-level binary dependency analysis


def _get_paths_for_parent_directory_preservation():
    """
    Return list of paths that serve as prefixes for parent-directory preservation of collected binaries and/or
    shared libraries. If a binary is collected from a location that starts with a path from this list, the relative
    directory structure is preserved within the frozen application bundle; otherwise, the binary is collected to the
    frozen application's top-level directory.
    """

    # Use only site-packages paths. We have no control over contents of `sys.path`, so using all paths from that may
    # lead to unintended behavior in corner cases. For example, if `sys.path` contained the drive root (see #7028),
    # all paths that do not match some other sub-path rooted in that drive will end up recognized as relative to the
    # drive root. In such case, any DLL collected from `c:\Windows\system32` will be collected into `Windows\system32`
    # sub-directory; ucrt DLLs collected from MSVC or Windows SDK installed in `c:\Program Files\...` will end up
    # collected into `Program Files\...` subdirectory; etc.
    #
    # On the other hand, the DLL parent directory preservation is primarily aimed at packages installed via PyPI
    # wheels, which are typically installed into site-packages. Therefore, limiting the directory preservation for
    # shared libraries collected from site-packages should do the trick, and should be reasonably safe.
    import site

    orig_paths = site.getsitepackages()
    orig_paths.append(site.getusersitepackages())

    # Explicitly excluded paths. `site.getsitepackages` seems to include `sys.prefix`, which we need to exclude, to
    # avoid issue swith DLLs in its sub-directories. We need both resolved and unresolved variant to handle cases
    # where `base_prefix` itself is a symbolic link (e.g., `scoop`-installed python on Windows, see #8023).
    excluded_paths = {
        pathlib.Path(sys.base_prefix),
        pathlib.Path(sys.base_prefix).resolve(),
        pathlib.Path(sys.prefix),
        pathlib.Path(sys.prefix).resolve(),
    }

    # For each path in orig_paths, append a resolved variant. This helps with linux venv where we need to consider
    # both `venv/lib/python3.11/site-packages` and `venv/lib/python3.11/site-packages` and `lib64` is a symlink
    # to `lib`.
    orig_paths += [pathlib.Path(path).resolve() for path in orig_paths]

    paths = set()
    for path in orig_paths:
        if not path:
            continue
        path = pathlib.Path(path)
        # Filter out non-directories (e.g., /path/to/python3x.zip) or non-existent paths
        if not path.is_dir():
            continue
        # Filter out explicitly excluded paths
        if path in excluded_paths:
            continue
        paths.add(path)

    # Sort by length (in term of path components) to ensure match against the longest common prefix (for example, match
    # /path/to/venv/lib/site-packages instead of /path/to/venv when both paths are in site paths).
    paths = sorted(paths, key=lambda x: len(x.parents), reverse=True)

    return paths


def _select_destination_directory(src_filename, parent_dir_preservation_paths):
    # Check parent directory preservation paths
    for parent_dir_preservation_path in parent_dir_preservation_paths:
        if parent_dir_preservation_path in src_filename.parents:
            # Collect into corresponding sub-directory.
            return src_filename.relative_to(parent_dir_preservation_path)

    # Collect into top-level directory.
    return src_filename.name


def binary_dependency_analysis(binaries, search_paths=None, symlink_suppression_patterns=None):
    """
    Perform binary dependency analysis on the given TOC list of collected binaries, by recursively scanning each binary
    for linked dependencies (shared library imports). Returns new TOC list that contains both original entries and their
    binary dependencies.

    Additional search paths for dependencies' full path resolution may be supplied via optional argument.
    """

    # Get all path prefixes for binaries' parent-directory preservation. For binaries collected from packages in (for
    # example) site-packages directory, we should try to preserve the parent directory structure.
    parent_dir_preservation_paths = _get_paths_for_parent_directory_preservation()

    # Keep track of processed binaries and processed dependencies.
    processed_binaries = set()
    processed_dependencies = set()

    # Keep track of unresolved dependencies, in order to defer the missing-library warnings until after everything has
    # been processed. This allows us to suppress warnings for dependencies that end up being collected anyway; for
    # details, see the end of this function.
    missing_dependencies = []

    # Populate output TOC with input binaries - this also serves as TODO list, as we iterate over it while appending
    # new entries at the end.
    output_toc = binaries[:]
    for dest_name, src_name, typecode in output_toc:
        # Do not process symbolic links (already present in input TOC list, or added during analysis below).
        if typecode == 'SYMLINK':
            continue

        # Keep track of processed binaries, to avoid unnecessarily repeating analysis of the same file. Use pathlib.Path
        # to avoid having to worry about case normalization.
        src_path = pathlib.Path(src_name)
        if src_path in processed_binaries:
            continue
        processed_binaries.add(src_path)

        logger.debug("Analyzing binary %r", src_name)

        # Analyze imports (linked dependencies)
        for dep_name, dep_src_path in get_imports(src_name, search_paths):
            logger.debug("Processing dependency, name: %r, resolved path: %r", dep_name, dep_src_path)

            # Skip unresolved dependencies. Defer the missing-library warnings until after binary dependency analysis
            # is complete.
            if not dep_src_path:
                missing_dependencies.append((dep_name, src_name))
                continue

            # Compare resolved dependency against global inclusion/exclusion rules.
            if not dylib.include_library(dep_src_path):
                logger.debug("Skipping dependency %r due to global exclusion rules.", dep_src_path)
                continue

            dep_src_path = pathlib.Path(dep_src_path)  # Turn into pathlib.Path for subsequent processing

            # Avoid processing this dependency if we have already processed it.
            if dep_src_path in processed_dependencies:
                logger.debug("Skipping dependency %r due to prior processing.", str(dep_src_path))
                continue
            processed_dependencies.add(dep_src_path)

            # Try to preserve parent directory structure, if applicable.
            # NOTE: do not resolve the source path, because on macOS and linux, it may be a versioned .so (e.g.,
            # libsomething.so.1, pointing at libsomething.so.1.2.3), and we need to collect it under original name!
            dep_dest_path = _select_destination_directory(dep_src_path, parent_dir_preservation_paths)
            dep_dest_path = pathlib.PurePath(dep_dest_path)  # Might be a str() if it is just a basename...

            # If we are collecting library into top-level directory on macOS, check whether it comes from a
            # .framework bundle. If it does, re-create the .framework bundle in the top-level directory
            # instead.
            if compat.is_darwin and dep_dest_path.parent == pathlib.PurePath('.'):
                if osxutils.is_framework_bundle_lib(dep_src_path):
                    # dst_src_path is parent_path/Name.framework/Versions/Current/Name
                    framework_parent_path = dep_src_path.parent.parent.parent.parent
                    dep_dest_path = pathlib.PurePath(dep_src_path.relative_to(framework_parent_path))

            logger.debug("Collecting dependency %r as %r.", str(dep_src_path), str(dep_dest_path))
            output_toc.append((str(dep_dest_path), str(dep_src_path), 'BINARY'))

            # On non-Windows, if we are not collecting the binary into application's top-level directory ('.'),
            # add a symbolic link from top-level directory to the actual location. This is to accommodate
            # LD_LIBRARY_PATH being set to the top-level application directory on linux (although library search
            # should be mostly done via rpaths, so this might be redundant) and to accommodate library path
            # rewriting on macOS, which assumes that the library was collected into top-level directory.
            if compat.is_win:
                # We do not use symlinks on Windows.
                pass
            elif dep_dest_path.parent == pathlib.PurePath('.'):
                # The shared library itself is being collected into top-level application directory.
                pass
            elif any(dep_src_path.match(pattern) for pattern in symlink_suppression_patterns):
                # Honor symlink suppression patterns specified by hooks.
                logger.debug(
                    "Skipping symbolic link from %r to top-level application directory due to source path matching one "
                    "of symlink suppression path patterns.", str(dep_dest_path)
                )
            else:
                logger.debug("Adding symbolic link from %r to top-level application directory.", str(dep_dest_path))
                output_toc.append((str(dep_dest_path.name), str(dep_dest_path), 'SYMLINK'))

    # Handle missing dependencies: display warnings, add missing symbolic links to top-level application directory, etc.
    seen_binaries = {
        os.path.normcase(os.path.basename(src_name)): (dest_name, src_name, typecode)
        for dest_name, src_name, typecode in output_toc if typecode != 'SYMLINK'
    }
    existing_symlinks = set([dest_name for dest_name, src_name, typecode in output_toc if typecode == 'SYMLINK'])

    for dependency_name, referring_binary in missing_dependencies:
        # Ignore libraries that we would not collect in the first place.
        if not dylib.include_library(dependency_name):
            continue

        # If the binary with a matching basename happens to be among the discovered binaries, suppress the message as
        # well. This might happen either because the library was collected by some other mechanism (for example, via
        # hook, or supplied by the user), or because it was discovered during the analysis of another binary (which,
        # for example, had properly set run-paths on Linux/macOS or was located next to that other analyzed binary on
        # Windows).
        #
        # On non-Windows, also check if symbolic link to the discovered binary already exists in the top-level
        # application directory, and if not, create it. This is important especially on macOS, where our library path
        # rewriting assumes that all dependent libraries are available in the top-level application directory, or
        # linked into it.
        dependency_basename = os.path.normcase(os.path.basename(dependency_name))
        dependency_toc_entry = seen_binaries.get(dependency_basename, None)
        if dependency_toc_entry is None:
            # Not found, emit a warning (subject to global warning suppression rules).
            if not dylib.warn_missing_lib(dependency_name):
                continue
            logger.warning(
                "Library not found: could not resolve %r, dependency of %r.", dependency_name, referring_binary
            )
        elif not compat.is_win:
            # Found; generate symbolic link if necessary.
            dependency_dest_path = pathlib.PurePath(dependency_toc_entry[0])
            dependency_src_path = pathlib.Path(dependency_toc_entry[1])

            if dependency_dest_path.parent == pathlib.PurePath('.'):
                # The binary is collected into top-level application directory.
                continue
            elif dependency_basename in existing_symlinks:
                # The symbolic link already exists.
                continue

            # Keep honoring symlink suppression patterns specified by hooks (same as in main binary dependency analysis
            # loop).
            if any(dependency_src_path.match(pattern) for pattern in symlink_suppression_patterns):
                logger.info(
                    "Missing dependency handling: skipping symbolic link from %r to top-level application directory "
                    "due to source path matching one of symlink suppression path patterns.", str(dependency_dest_path)
                )
                continue

            # Create the symbolic link
            logger.info(
                "Missing dependency handling: adding symbolic link from %r to top-level application directory.",
                str(dependency_dest_path)
            )
            output_toc.append((dependency_basename, str(dependency_dest_path), 'SYMLINK'))
            existing_symlinks.add(dependency_basename)

    return output_toc


#- Low-level import analysis


def get_imports(filename, search_paths=None):
    """
    Analyze the given binary file (shared library or executable), and obtain the list of shared libraries it imports
    (i.e., link-time dependencies).

    Returns set of tuples (name, fullpath). The name component is the referenced name, and on macOS, may not be just
    a base name. If the library's full path cannot be resolved, fullpath element is None.

    Additional list of search paths may be specified via `search_paths`, to be used as a fall-back when the
    platform-specific resolution mechanism fails to resolve a library fullpath.
    """
    if compat.is_win:
        if str(filename).lower().endswith(".manifest"):
            return []
        return _get_imports_pefile(filename, search_paths)
    elif compat.is_darwin:
        return _get_imports_macholib(filename, search_paths)
    else:
        return _get_imports_ldd(filename, search_paths)


def _get_imports_pefile(filename, search_paths):
    """
    Windows-specific helper for `get_imports`, which uses the `pefile` library to walk through PE header.
    """
    import pefile

    output = set()

    # By default, pefile library parses all PE information. We are only interested in the list of dependent dlls.
    # Performance is improved by reading only needed information. https://code.google.com/p/pefile/wiki/UsageExamples
    pe = pefile.PE(filename, fast_load=True)
    pe.parse_data_directories(
        directories=[
            pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_IMPORT'],
            pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_EXPORT'],
        ],
        forwarded_exports_only=True,
        import_dllnames_only=True,
    )

    # If a library has no binary dependencies, pe.DIRECTORY_ENTRY_IMPORT does not exist.
    for entry in getattr(pe, 'DIRECTORY_ENTRY_IMPORT', []):
        dll_str = entry.dll.decode('utf-8')
        output.add(dll_str)

    # We must also read the exports table to find forwarded symbols:
    # http://blogs.msdn.com/b/oldnewthing/archive/2006/07/19/671238.aspx
    exported_symbols = getattr(pe, 'DIRECTORY_ENTRY_EXPORT', None)
    if exported_symbols:
        for symbol in exported_symbols.symbols:
            if symbol.forwarder is not None:
                # symbol.forwarder is a bytes object. Convert it to a string.
                forwarder = symbol.forwarder.decode('utf-8')
                # symbol.forwarder is for example 'KERNEL32.EnterCriticalSection'
                dll = forwarder.split('.')[0]
                output.add(dll + ".dll")

    pe.close()

    # Attempt to resolve full paths to referenced DLLs. Always add the input binary's parent directory to the search
    # paths.
    search_paths = [os.path.dirname(filename)] + (search_paths or [])
    output = {(lib, resolve_library_path(lib, search_paths)) for lib in output}

    return output


def _get_imports_ldd(filename, search_paths):
    """
    Helper for `get_imports`, which uses `ldd` to analyze shared libraries. Used on Linux and other POSIX-like platforms
    (with exception of macOS).
    """

    output = set()

    # Output of ldd varies between platforms...
    if compat.is_aix:
        # Match libs of the form
        #   'archivelib.a(objectmember.so/.o)'
        # or
        #   'sharedlib.so'
        # Will not match the fake lib '/unix'
        LDD_PATTERN = re.compile(r"^\s*(((?P<libarchive>(.*\.a))(?P<objectmember>\(.*\)))|((?P<libshared>(.*\.so))))$")
    elif compat.is_hpux:
        # Match libs of the form
        #   'sharedlib.so => full-path-to-lib
        # e.g.
        #   'libpython2.7.so =>      /usr/local/lib/hpux32/libpython2.7.so'
        LDD_PATTERN = re.compile(r"^\s+(.*)\s+=>\s+(.*)$")
    elif compat.is_solar:
        # Match libs of the form
        #   'sharedlib.so => full-path-to-lib
        # e.g.
        #   'libpython2.7.so.1.0 => /usr/local/lib/libpython2.7.so.1.0'
        # Will not match the platform specific libs starting with '/platform'
        LDD_PATTERN = re.compile(r"^\s+(.*)\s+=>\s+(.*)$")
    elif compat.is_linux:
        # Match libs of the form
        #   libpython3.13.so.1.0 => /home/brenainn/.pyenv/versions/3.13.0/lib/libpython3.13.so.1.0 (0x00007a9e15800000)
        # or
        #   /tmp/python/install/bin/../lib/libpython3.13.so.1.0 (0x00007b9489c82000)
        LDD_PATTERN = re.compile(r"^\s*(?:(.*?)\s+=>\s+)?(.*?)\s+\(.*\)")
    else:
        LDD_PATTERN = re.compile(r"\s*(.*?)\s+=>\s+(.*?)\s+\(.*\)")

    # Resolve symlinks since GNU ldd contains a bug in processing a symlink to a binary
    # using $ORIGIN: https://sourceware.org/bugzilla/show_bug.cgi?id=25263
    p = subprocess.run(
        ['ldd', os.path.realpath(filename)],
        stdin=subprocess.DEVNULL,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        encoding='utf-8',
    )

    ldd_warnings = []
    for line in p.stderr.splitlines():
        if not line:
            continue
        # Python extensions (including stdlib ones) are not linked against python.so but rely on Python's symbols having
        # already been loaded into symbol space at runtime. musl's ldd issues a series of harmless warnings to stderr
        # telling us that those symbols are unfindable. These should be suppressed.
        elif line.startswith("Error relocating ") and line.endswith(" symbol not found"):
            continue
        # Shared libraries should have the executable bits set; however, this is not the case for shared libraries
        # shipped in PyPI wheels, which cause ldd to emit `ldd: warning: you do not have execution permission for ...`
        # warnings. Suppress these.
        elif line.startswith("ldd: warning: you do not have execution permission for "):
            continue
        # When `ldd` is ran against a file that is not a dynamic binary (i.e., is not a binary at all, or is a static
        # binary), it emits a "not a dynamic executable" warning. Suppress it.
        elif "not a dynamic executable" in line:
            continue
        # Propagate any other warnings it might have.
        ldd_warnings.append(line)
    if ldd_warnings:
        logger.warning("ldd warnings for %r:\n%s", filename, "\n".join(ldd_warnings))

    for line in p.stdout.splitlines():
        name = None  # Referenced name
        lib = None  # Resolved library path

        m = LDD_PATTERN.search(line)
        if m:
            if compat.is_aix:
                libarchive = m.group('libarchive')
                if libarchive:
                    # We matched an archive lib with a request for a particular embedded shared object.
                    #   'archivelib.a(objectmember.so/.o)'
                    lib = libarchive
                    name = os.path.basename(lib) + m.group('objectmember')
                else:
                    # We matched a stand-alone shared library.
                    #   'sharedlib.so'
                    lib = m.group('libshared')
                    name = os.path.basename(lib)
            elif compat.is_hpux:
                name, lib = m.group(1), m.group(2)
            else:
                name, lib = m.group(1), m.group(2)
                name = name or os.path.basename(lib)
                if compat.is_linux:
                    # Skip all ld variants listed https://sourceware.org/glibc/wiki/ABIList
                    # plus musl's ld-musl-*.so.*.
                    if re.fullmatch(r"ld(64)?(-linux|-musl)?(-.+)?\.so(\..+)?", os.path.basename(lib)):
                        continue
            if name[:10] in ('linux-gate', 'linux-vdso'):
                # linux-gate is a fake library which does not exist and should be ignored. See also:
                # http://www.trilithium.com/johan/2005/08/linux-gate/
                continue

            if compat.is_cygwin:
                # exclude Windows system library
                if lib.lower().startswith('/cygdrive/c/windows/system'):
                    continue

            # Reset library path if it does not exist
            if not os.path.exists(lib):
                lib = None
        elif line.endswith("not found"):
            # On glibc-based linux distributions, missing libraries are marked with name.so => not found
            tokens = line.split('=>')
            if len(tokens) != 2:
                continue
            name = tokens[0].strip()
            lib = None
        else:
            # TODO: should we warn about unprocessed lines?
            continue

        # Fall back to searching the supplied search paths, if any.
        if not lib:
            lib = _resolve_library_path_in_search_paths(
                os.path.basename(name),  # Search for basename of the referenced name.
                search_paths,
            )

        # Normalize the resolved path, to remove any extraneous "../" elements.
        if lib:
            lib = os.path.normpath(lib)

        # Return referenced name as-is instead of computing a basename, to provide additional context when library
        # cannot be resolved.
        output.add((name, lib))

    return output


def _get_imports_macholib(filename, search_paths):
    """
    macOS-specific helper for `get_imports`, which uses `macholib` to analyze library load commands in Mach-O headers.
    """
    from macholib.dyld import dyld_find
    from macholib.mach_o import LC_RPATH
    from macholib.MachO import MachO

    try:
        from macholib.dyld import _dyld_shared_cache_contains_path
    except ImportError:
        _dyld_shared_cache_contains_path = None

    output = set()

    # Parent directory of the input binary and parent directory of python executable, used to substitute @loader_path
    # and @executable_path. The macOS dylib loader (dyld) fully resolves the symbolic links when using @loader_path
    # and @executable_path references, so we need to do the same using `os.path.realpath`.
    bin_path = os.path.dirname(os.path.realpath(filename))
    python_bin = os.path.realpath(sys.executable)
    python_bin_path = os.path.dirname(python_bin)

    def _get_referenced_libs(m):
        # Collect referenced libraries from MachO object.
        referenced_libs = set()
        for header in m.headers:
            for idx, name, lib in header.walkRelocatables():
                referenced_libs.add(lib)
        return referenced_libs

    def _get_run_paths(m):
        # Find LC_RPATH commands to collect rpaths from MachO object.
        # macholib does not handle @rpath, so we need to handle run paths ourselves.
        run_paths = []
        for header in m.headers:
            for command in header.commands:
                # A command is a tuple like:
                #   (<macholib.mach_o.load_command object at 0x>,
                #    <macholib.mach_o.rpath_command object at 0x>,
                #    '../lib\x00\x00')
                cmd_type = command[0].cmd
                if cmd_type == LC_RPATH:
                    rpath = command[2].decode('utf-8')
                    # Remove trailing '\x00' characters. E.g., '../lib\x00\x00'
                    rpath = rpath.rstrip('\x00')
                    # If run path starts with @, ensure it starts with either @loader_path or @executable_path.
                    # We cannot process anything else.
                    if rpath.startswith("@") and not rpath.startswith(("@executable_path", "@loader_path")):
                        logger.warning("Unsupported rpath format %r found in binary %r - ignoring...", rpath, filename)
                        continue
                    run_paths.append(rpath)
        return run_paths

    @functools.lru_cache
    def get_run_paths_and_referenced_libs(filename):
        # Walk through Mach-O headers, and collect all referenced libraries and run paths.
        m = MachO(filename)
        return _get_referenced_libs(m), _get_run_paths(m)

    @functools.lru_cache
    def get_run_paths(filename):
        # Walk through Mach-O headers, and collect only run paths.
        return _get_run_paths(MachO(filename))

    # Collect referenced libraries and run paths from the input binary.
    referenced_libs, run_paths = get_run_paths_and_referenced_libs(filename)

    # On macOS, run paths (rpaths) are inherited from the executable that loads the given shared library (or from the
    # shared library that loads the given shared library). This means that shared libraries and python binary extensions
    # can reference other shared libraries using @rpath without having set any run paths themselves.
    #
    # In order to simulate the run path inheritance that happens in unfrozen python programs, we need to augment the
    # run paths from the given binary with those set by the python interpreter executable (`sys.executable`). Anaconda
    # python, for example, sets the run path on the python executable to `@loader_path/../lib`, which allows python
    # extensions to reference shared libraries in the Anaconda environment's `lib` directory via only `@rpath`
    # (for example, the `_ssl` extension can reference the OpenSSL library as `@rpath/libssl.3.dylib`). In another
    # example, python executable has its run path set to the top-level directory of its .framework bundle; in this
    # case the `ssl` extension references the OpenSSL library as `@rpath/Versions/3.10/lib/libssl.1.1.dylib`.
    run_paths += get_run_paths(python_bin)

    # This fallback should be fully superseded by the above recovery of run paths from python executable; but for now,
    # keep it around in case of unforeseen corner cases.
    run_paths.append(os.path.join(compat.base_prefix, 'lib'))

    # De-duplicate run_paths while preserving their order.
    run_paths = list(dict.fromkeys(run_paths))

    def _resolve_using_path(lib):
        # Absolute paths should not be resolved; we should just check whether the library exists or not. This used to
        # be done using macholib's dyld_find() as well (as it properly handles system libraries that are hidden on
        # Big Sur and later), but it turns out that even if given an absolute path, it gives precedence to search paths
        # from DYLD_LIBRARY_PATH. This leads to confusing errors when directory in DYLD_LIBRARY_PATH contains a file
        # (shared library or data file) that happens to have the same name as a library from a system framework.
        if os.path.isabs(lib):
            if _dyld_shared_cache_contains_path is not None and _dyld_shared_cache_contains_path(lib):
                return lib
            if os.path.isfile(lib):
                return lib
            return None

        try:
            return dyld_find(lib)
        except ValueError:
            return None

    def _resolve_using_loader_path(lib, bin_path, python_bin_path):
        # Strictly speaking, @loader_path should be anchored to parent directory of analyzed binary (`bin_path`), while
        # @executable_path should be anchored to the parent directory of the process' executable. Typically, this would
        # be python executable (`python_bin_path`). Unless we are analyzing a collected 3rd party executable; in that
        # case, `bin_path` is correct option. So we first try resolving using `bin_path`, and then fall back to
        # `python_bin_path`. This does not account for transitive run paths of higher-order dependencies, but there is
        # only so much we can do here...
        #
        # NOTE: do not use macholib's `dyld_find`, because its fallback search locations might end up resolving wrong
        # instance of the library! For example, if our `bin_path` and `python_bin_path` are anchored in an Anaconda
        # python environment and the candidate library path does not exit (because we are calling this function when
        # trying to resolve @rpath with multiple candidate run paths), we do not want to fall back to eponymous library
        # that happens to be present in the Homebrew python environment...
        if lib.startswith('@loader_path/'):
            lib = lib[len('@loader_path/'):]
        elif lib.startswith('@executable_path/'):
            lib = lib[len('@executable_path/'):]

        # Try resolving with binary's path first...
        resolved_lib = _resolve_using_path(os.path.join(bin_path, lib))
        if resolved_lib is not None:
            return resolved_lib

        # ... and fall-back to resolving with python executable's path
        return _resolve_using_path(os.path.join(python_bin_path, lib))

    # Try to resolve full path of the referenced libraries.
    for referenced_lib in referenced_libs:
        resolved_lib = None

        # If path starts with @rpath, we have to handle it ourselves.
        if referenced_lib.startswith('@rpath'):
            lib = os.path.join(*referenced_lib.split(os.sep)[1:])  # Remove the @rpath/ prefix

            # Try all run paths.
            for run_path in run_paths:
                # Join the path.
                lib_path = os.path.join(run_path, lib)

                if lib_path.startswith(("@executable_path", "@loader_path")):
                    # Run path starts with @executable_path or @loader_path.
                    lib_path = _resolve_using_loader_path(lib_path, bin_path, python_bin_path)
                else:
                    # If run path was relative, anchor it to binary's location.
                    if not os.path.isabs(lib_path):
                        os.path.join(bin_path, lib_path)
                    lib_path = _resolve_using_path(lib_path)

                if lib_path and os.path.exists(lib_path):
                    resolved_lib = lib_path
                    break
        else:
            if referenced_lib.startswith(("@executable_path", "@loader_path")):
                resolved_lib = _resolve_using_loader_path(referenced_lib, bin_path, python_bin_path)
            else:
                resolved_lib = _resolve_using_path(referenced_lib)

        # Fall back to searching the supplied search paths, if any.
        if not resolved_lib:
            resolved_lib = _resolve_library_path_in_search_paths(
                os.path.basename(referenced_lib),  # Search for basename of the referenced name.
                search_paths,
            )

        # Normalize the resolved path, to remove any extraneous "../" elements.
        if resolved_lib:
            resolved_lib = os.path.normpath(resolved_lib)

        # Return referenced library name as-is instead of computing a basename. Full referenced name carries additional
        # information that might be useful for the caller to determine how to deal with unresolved library (e.g., ignore
        # unresolved libraries that are supposed to be located in system-wide directories).
        output.add((referenced_lib, resolved_lib))

    return output


#- Library full path resolution


def resolve_library_path(name, search_paths=None):
    """
    Given a library name, attempt to resolve full path to that library. The search for library is done via
    platform-specific mechanism and fall back to optionally-provided list of search paths. Returns None if library
    cannot be resolved. If give library name is already an absolute path, the given path is returned without any
    processing.
    """
    # No-op if path is already absolute.
    if os.path.isabs(name):
        return name

    if compat.is_unix:
        # Use platform-specific helper.
        fullpath = _resolve_library_path_unix(name)
        if fullpath:
            return fullpath
        # Fall back to searching the supplied search paths, if any
        return _resolve_library_path_in_search_paths(name, search_paths)
    elif compat.is_win:
        # Try the caller-supplied search paths, if any.
        fullpath = _resolve_library_path_in_search_paths(name, search_paths)
        if fullpath:
            return fullpath

        # Fall back to default Windows search paths, using the PATH environment variable (which should also include
        # the system paths, such as c:\windows and c:\windows\system32)
        win_search_paths = [path for path in compat.getenv('PATH', '').split(os.pathsep) if path]
        return _resolve_library_path_in_search_paths(name, win_search_paths)
    else:
        return ctypes.util.find_library(name)

    return None


# Compatibility aliases for hooks from contributed hooks repository. All of these now point to the high-level
# `resolve_library_path`.
findLibrary = resolve_library_path
findSystemLibrary = resolve_library_path


def _resolve_library_path_in_search_paths(name, search_paths=None):
    """
    Low-level helper for resolving given library name to full path in given list of search paths.
    """
    for search_path in search_paths or []:
        fullpath = os.path.join(search_path, name)
        if not os.path.isfile(fullpath):
            continue

        # On Windows, ensure that architecture matches that of running python interpreter.
        if compat.is_win:
            try:
                dll_machine_type = winutils.get_pe_file_machine_type(fullpath)
            except Exception:
                # A search path might contain a DLL that we cannot analyze; for example, a stub file. Skip over.
                continue
            if dll_machine_type != _exe_machine_type:
                continue

        return os.path.normpath(fullpath)

    return None


def _resolve_library_path_unix(name):
    """
    UNIX-specific helper for resolving library path.

    Emulates the algorithm used by dlopen. `name` must include the prefix, e.g., ``libpython2.4.so``.
    """
    assert compat.is_unix, "Current implementation for Unix only (Linux, Solaris, AIX, FreeBSD)"

    if name.endswith('.so') or '.so.' in name:
        # We have been given full library name that includes suffix. Use `_resolve_library_path_in_search_paths` to find
        # the exact match.
        lib_search_func = _resolve_library_path_in_search_paths
    else:
        # We have been given a library name without suffix. Use `_which_library` as search function, which will try to
        # find library with matching basename.
        lib_search_func = _which_library

    # Look in the LD_LIBRARY_PATH according to platform.
    if compat.is_aix:
        lp = compat.getenv('LIBPATH', '')
    elif compat.is_darwin:
        lp = compat.getenv('DYLD_LIBRARY_PATH', '')
    else:
        lp = compat.getenv('LD_LIBRARY_PATH', '')
    lib = lib_search_func(name, filter(None, lp.split(os.pathsep)))

    # Look in /etc/ld.so.cache
    # Solaris does not have /sbin/ldconfig. Just check if this file exists.
    if lib is None:
        utils.load_ldconfig_cache()
        lib = utils.LDCONFIG_CACHE.get(name)
        if lib:
            assert os.path.isfile(lib)

    # Look in the known safe paths.
    if lib is None:
        # Architecture independent locations.
        paths = ['/lib', '/usr/lib']
        # Architecture dependent locations.
        if compat.architecture == '32bit':
            paths.extend(['/lib32', '/usr/lib32'])
        else:
            paths.extend(['/lib64', '/usr/lib64'])
        # Machine dependent locations.
        if compat.machine == 'intel':
            if compat.architecture == '32bit':
                paths.extend(['/usr/lib/i386-linux-gnu'])
            else:
                paths.extend(['/usr/lib/x86_64-linux-gnu'])

        # On Debian/Ubuntu /usr/bin/python is linked statically with libpython. Newer Debian/Ubuntu with multiarch
        # support puts the libpythonX.Y.so in paths like /usr/lib/i386-linux-gnu/. Try to query the arch-specific
        # sub-directory, if available.
        arch_subdir = sysconfig.get_config_var('multiarchsubdir')
        if arch_subdir:
            arch_subdir = os.path.basename(arch_subdir)
            paths.append(os.path.join('/usr/lib', arch_subdir))
        else:
            logger.debug('Multiarch directory not detected.')

        # Termux (a Ubuntu like subsystem for Android) has an additional libraries directory.
        if os.path.isdir('/data/data/com.termux/files/usr/lib'):
            paths.append('/data/data/com.termux/files/usr/lib')

        if compat.is_aix:
            paths.append('/opt/freeware/lib')
        elif compat.is_hpux:
            if compat.architecture == '32bit':
                paths.append('/usr/local/lib/hpux32')
            else:
                paths.append('/usr/local/lib/hpux64')
        elif compat.is_freebsd or compat.is_openbsd:
            paths.append('/usr/local/lib')
        lib = lib_search_func(name, paths)

    return lib


def _which_library(name, dirs):
    """
    Search for a shared library in a list of directories.

    Args:
        name:
            The library name including the `lib` prefix but excluding any `.so` suffix.
        dirs:
            An iterable of folders to search in.
    Returns:
        The path to the library if found or None otherwise.

    """
    matcher = _library_matcher(name)
    for path in filter(os.path.exists, dirs):
        for _path in os.listdir(path):
            if matcher(_path):
                return os.path.join(path, _path)


def _library_matcher(name):
    """
    Create a callable that matches libraries if **name** is a valid library prefix for input library full names.
    """
    return re.compile(name + r"[0-9]*\.").match


#- Python shared library search


def get_python_library_path():
    """
    Find Python shared library that belongs to the current interpreter.

    Return  full path to Python dynamic library or None when not found.

    PyInstaller needs to collect the Python shared library, so that bootloader can load it, import Python C API
    symbols, and use them to set up the embedded Python interpreter.

    The name of the shared library is typically fixed (`python3.X.dll` on Windows, libpython3.X.so on Unix systems,
    and `libpython3.X.dylib` on macOS for shared library builds and `Python.framework/Python` for framework build).
    Its location can usually be inferred from the Python interpreter executable, when the latter is dynamically
    linked against the shared library.

    However, some situations require extra handling due to various quirks; for example, Debian-based linux
    distributions statically link the Python interpreter executable against the Python library, while also providing
    a shared library variant for external users.
    """

    # With Windows Python builds, this is pretty straight-forward: `sys.dllhandle` provides a handle to the loaded
    # Python DLL, and we can resolve its path using `GetModuleFileName()` from win32 API.
    # This is applicable to python.org Windows builds, Anaconda on Windows, and MSYS2 Python.
    if compat.is_win:
        if hasattr(sys, 'dllhandle'):
            import _winapi
            return _winapi.GetModuleFileName(sys.dllhandle)
        else:
            raise PythonLibraryNotFoundError(
                "Python was built without a shared library, which is required by PyInstaller."
            )

    # On other (POSIX) platforms, the name of the Python shared library is available in the `INSTSONAME` variable
    # exposed by the `sysconfig` module. There is also the `LDLIBRARY` variable, which points to the unversioned .so
    # symbolic link for linking purposes; however, we are interested in the actual, fully-versioned soname.
    # This should cover all variations in the naming schemes across different platforms as well as different build
    # options (debug build, free-threaded build, etc.).
    #
    # However, `INSTSONAME` points to the shared library only if shared library is enabled; in static-library builds,
    # it points to the static library, which is of no use to us. We can check if Python was built with shared library
    # (i.e., the `--enable-shared` option) by checking `Py_ENABLE_SHARED` variable, which should be set to 1 in this
    # case (and 0 in the case of a static-library build). On macOS, builds made with `--enable-framework` have
    # `Py_ENABLE_SHARED` set to 0, but have `PYTHONFRAMEWORK`set to a non-empty string.
    #
    # The above description is further complicated by the fact that in some Python builds, the `python` executable is
    # built against static Python library, and the shared library is built separately and provided for development and
    # for embedders (such as PyInstaller). Presumably, this is done for performance reasons. Also, it is enabled by the
    # fact that on POSIX, Python extensions do not need to have the referenced Python symbols resolved at link-time;
    # rather, these symbols can be resolved at run-time from the running Python process (and are effectively provided
    # by the `python` executable). Such builds come in two variants. In the first variant, `Py_ENABLE_SHARED` is 0 and
    # `INSTSONAME` points to the static library; an example of such build is Anaconda Python. In the second variant,
    # `Py_ENABLE_SHARED` is 1 and `INSTSONAME` points to the shared library, but `python` executable is not linked
    # against it; examples of such build are Debian-packaged Python and `astral-sh/python-build-standalone` Python.
    #
    # Therefore, our strategy is as follows: if we determine that shared library was enabled (via `Py_ENABLE_SHARED`
    # on all platforms and/or via `PYTHONFRAMEWORK` on macOS), we use the name given by `INSTSONAME`. First, we try
    # to locate it by analyzing binary dependencies of `python` executable (regular shared-library-enabled build),
    # then fall back to standard search locations (second variant of static-executable-with-separate-shared-library).
    # If `Py_ENABLE_SHARED` is set to 0, we try to guess the library name based on version and feature flags, but we
    # search only `sys.base_prefix` and `lib` directory under `sys.base_prefix`; if the shared library is not found
    # there, we assume it is unavailable and raise an error. This attempts to accommodate Anaconda python (and corner
    # cases when we cannot reliably identify Anaconda python - see #9273) and prevent accidental bundling of
    # system-wide Python shared library in cases when user tries to use custom Python build without shared library.

    def _find_lib_in_libdirs(name, *libdirs):
        for libdir in libdirs:
            full_path = os.path.join(libdir, name)
            if not os.path.exists(full_path):
                continue
            # Resolve potential symbolic links to achieve consistent results with linker-based search; e.g., on
            # POSIX systems, linker resolves unversioned library names (python3.X.so) to versioned ones
            # (libpython3.X.so.1.0) due to former being symbolic links to the latter. See #6831.
            full_path = os.path.realpath(full_path)
            if not os.path.exists(full_path):
                continue
            return full_path
        return None

    is_shared = (
        # Builds made with `--enable-shared` have `Py_ENABLE_SHARED` set to 1. This is true even for Debian-packaged
        # Python, which has the `python` executable statically linked against the Python library.
        sysconfig.get_config_var("Py_ENABLE_SHARED") or
        # On macOS, builds made with `--enable-framework` have `Py_ENABLE_SHARED` set to 0, but have `PYTHONFRAMEWORK`
        # set to a non-empty string.
        (compat.is_darwin and sysconfig.get_config_var("PYTHONFRAMEWORK"))
    )

    if not is_shared:
        # Anaconda Python; this codepath used to be under `compat.is_conda` switch, but we may also be dealing with
        # Anaconda Python without `conda-meta` directory (see #9273). Or some other Python build where shared library
        # is provided but `Py_ENABLE_SHARED` is set to 0.
        py_major, py_minor = sys.version_info[:2]
        py_suffix = "t" if compat.is_nogil else ""  # TODO: does Anaconda provide debug builds with "d" suffix?
        if compat.is_darwin:
            # macOS
            expected_name = f"libpython{py_major}.{py_minor}{py_suffix}.dylib"
        else:
            # Linux; assume any other potential POSIX builds use the same naming scheme.
            expected_name = f"libpython{py_major}.{py_minor}{py_suffix}.so.1.0"

        # Allow the library to be only in `sys.base_prefix` or the `lib` directory under it. This should prevent us from
        # picking up an unrelated copy of shared library that might happen to be available in standard search path, when
        # we should instead be raising an error due to Python having been built without a shared library. (In true
        # static-library builds, Python's own extension modules are usually turned into built-ins. So picking up an
        # unrelated Python shared library that happens to be of the same version results in run-time errors due to
        # missing extensions - because in the build that produced the shared library, those extensions are expected to
        # be external extension modules!)
        python_libname = _find_lib_in_libdirs(
            expected_name,  # Full name
            compat.base_prefix,
            os.path.join(compat.base_prefix, 'lib'),
        )
        if python_libname:
            return python_libname

        # Raise PythonLibraryNotFoundError
        option_str = (
            "either the `--enable-shared` or the `--enable-framework` option"
            if compat.is_darwin else "the `--enable-shared` option"
        )
        raise PythonLibraryNotFoundError(
            "Python was built without a shared library, which is required by PyInstaller. "
            f"If you built Python from source, rebuild it with {option_str}."
        )

    # Use the library name from `INSTSONAME`.
    expected_name = sysconfig.get_config_var('INSTSONAME')

    # In Cygwin builds (and also MSYS2 python, although that should be handled by Windows-specific codepath...),
    # INSTSONAME is available, but the name has a ".dll.a" suffix; remove that trailing ".a".
    if (compat.is_win or compat.is_cygwin) and os.path.normcase(expected_name).endswith('.dll.a'):
        expected_name = expected_name[:-2]

    # NOTE: on macOS with .framework bundle build, INSTSONAME contains full name of the .framework library, for example
    # `Python.framework/Versions/3.13/Python`. Pre-compute a basename for comparisons that are using only basename.
    expected_basename = os.path.normcase(os.path.basename(expected_name))

    # First, try to find the expected name among the libraries against which the Python executable is linked. This
    # assumes that the Python executable was not statically linked against the library (as is the case with
    # Debian-packaged Python or `astral-sh/python-build-standalone` Python).
    imported_libraries = get_imports(compat.python_executable)  # (name, fullpath) tuples
    for _, lib_path in imported_libraries:
        if lib_path is None:
            continue  # Skip unresolved imports
        if os.path.normcase(os.path.basename(lib_path)) == expected_basename:  # Basename comparison
            # Python library found. Return absolute path to it.
            return lib_path

    # As a fallback, try to find the library in several "standard" search locations...

    # First, search the `sys.base_prefix` and `lib` directory in `sys.base_prefix`, as these locations have the closest
    # ties to our current Python process. This caters to builds such as `astral-sh/python-build-standalone` Python.
    python_libname = _find_lib_in_libdirs(
        expected_name,  # Full name
        compat.base_prefix,
        os.path.join(compat.base_prefix, 'lib'),
    )
    if python_libname:
        return python_libname

    # Perform search in the configured library search locations. This should be done after exhausting all other options;
    # it primarily caters to Debian-packaged Python, but we need to make sure that we do not collect shared library from
    # system-installed Python when the current interpreter is in fact some other Python build (such as, for example,
    # `astral-sh/python-build-standalone` Python that is handled in the preceding code block).
    python_libname = resolve_library_path(expected_basename)  # Basename
    if python_libname:
        return python_libname

    # Not found. Raise a PythonLibraryNotFoundError with corresponding message.
    message = f"ERROR: Python shared library ({expected_name!r}) was not found!"
    if compat.is_linux and os.path.isfile('/etc/debian_version'):
        # The shared library is provided by `libpython3.x` package (i.e., no need to install full `python3-dev`).
        pkg_name = f"libpython3.{sys.version_info.minor}"
        message += (
            " If you are using system python on Debian/Ubuntu, you might need to install a separate package by running "
            f"`apt install {pkg_name}`."
        )

    raise PythonLibraryNotFoundError(message)


#- Binary vs data (re)classification


def classify_binary_vs_data(filename):
    """
    Classify the given file as either BINARY or a DATA, using appropriate platform-specific method. Returns 'BINARY'
    or 'DATA' string depending on the determined file type, or None if classification cannot be performed (non-existing
    file, missing tool, and other errors during classification).
    """

    # We cannot classify non-existent files.
    if not os.path.isfile(filename):
        return None

    # Use platform-specific implementation.
    return _classify_binary_vs_data(filename)


if compat.is_linux:

    def _classify_binary_vs_data(filename):
        # First check for ELF signature, in order to avoid calling `objdump` on every data file, which can be costly.
        try:
            with open(filename, 'rb') as fp:
                sig = fp.read(4)
        except Exception:
            return None

        if sig != b"\x7FELF":
            return "DATA"

        # Verify the binary by checking if `objdump` recognizes the file. The preceding ELF signature check should
        # ensure that this is an ELF file, while this check should ensure that it is a valid ELF file. In the future,
        # we could try checking that the architecture matches the running platform.
        cmd_args = ['objdump', '-a', filename]
        try:
            p = subprocess.run(
                cmd_args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.DEVNULL,
                encoding='utf8',
            )
        except Exception:
            return None  # Failed to run `objdump` or `objdump` unavailable.

        return 'BINARY' if p.returncode == 0 else 'DATA'

elif compat.is_win:

    @functools.lru_cache()
    def _no_op_pefile_gc():
        # Disable pefile's reduntant and very slow call to gc.collect(). See #8762.
        import types
        import gc
        import pefile

        fake_gc = types.ModuleType("gc")
        fake_gc.__dict__.update(gc.__dict__)
        fake_gc.collect = lambda *_, **__: None
        pefile.gc = fake_gc

    def _classify_binary_vs_data(filename):
        import pefile

        _no_op_pefile_gc()

        # First check for MZ signature, which should allow us to quickly classify the majority of data files.
        try:
            with open(filename, 'rb') as fp:
                sig = fp.read(2)
        except Exception:
            return None

        if sig != b"MZ":
            return "DATA"

        # Check if the file can be opened using `pefile`.
        try:
            with pefile.PE(filename, fast_load=True) as pe:  # noqa: F841
                pass
            return 'BINARY'
        except pefile.PEFormatError:
            return 'DATA'
        except Exception:
            pass

        return None

elif compat.is_darwin:

    def _classify_binary_vs_data(filename):
        # See if the file can be opened using `macholib`.
        import macholib.MachO

        try:
            macho = macholib.MachO.MachO(filename)  # noqa: F841
            return 'BINARY'
        except Exception:
            # TODO: catch only `ValueError`?
            pass

        return 'DATA'

else:

    def _classify_binary_vs_data(filename):
        # Classification not implemented for the platform.
        return None



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Mock package defining and then deleting a global variable of the same name as a mock submodule of this package.

This package is exercised by the `test_import_submodule_global_unshadowed` functional test.
"""

submodule = 'And with strange aeons even death may die.'
"""
Global variable of the same name as a mock submodule of this package.

This variable's value is both arbitrary _and_ always ignored by this test.
"""

# Permit the "submodule" submodule to be imported. Since globals take precedence over submodules of the same name,
# failing to undefine this global would prevent this submodule from being imported.
#
# PyInstaller explicitly detects both the definition and undefinition of globals in the modules containing those
# globals -- namely, here. PyInstaller does not, however, detect either the definition or undefinition of these globals
# from other modules. In particular, PyInstaller ignores attempts to undefine this global from the functional test
# exercising this module (e.g., using "del pyi_testmod_submodule_global.submodule").
del submodule



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# Ensure that environment variables TCL_LIBRARY and TK_LIBRARY are set properly, and that data files are collected.
# NOTE: "library" here refers to the scripts directory as in "collection", not as a dynamic/shared library.

# NOTE: on macOS, we do collect Tcl/Tk files when the _tkinter module is linked against system copy of Tcl/Tk framework.
# In that case, TCL_LIBRARY and TK_LIBRARY environment variables are not set by the runtime hook, and this test is
# reduced to a basic "import tkinter" test.

import glob
import os
import sys

import tkinter  # noqa: F401


def compare(test_name, expect, frozen):
    expect = os.path.normpath(expect)
    frozen = os.path.normpath(frozen)
    print(test_name)
    print(('  Expected: ' + expect))
    print(('  Current:  ' + frozen))
    print('')
    # Path must match.
    if not frozen == expect:
        raise SystemExit('Data directory is not set properly.')
    # Directory must exist.
    if not os.path.exists(frozen):
        raise SystemExit('Data directory does not exist.')
    # Directory must contain some .tcl files and not to be empty.
    if not len(glob.glob(frozen + '/*.tcl')) > 0:
        raise SystemExit('Data directory does not contain .tcl files.')


# Tcl scripts directory
tcl_dir = os.environ.get('TCL_LIBRARY')
if tcl_dir:
    compare('Tcl', os.path.join(sys.prefix, '_tcl_data'), tcl_dir)
elif sys.platform != 'darwin':
    raise SystemExit("TCL_LIBRARY environment variable is not set!")

# Tk scripts directory
tk_dir = os.environ.get('TK_LIBRARY')
if tk_dir:
    compare('Tk', os.path.join(sys.prefix, '_tk_data'), tk_dir)
elif sys.platform != 'darwin':
    raise SystemExit("TK_LIBRARY environment variable is not set!")
